# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file defines the behavior of agents."""
import random
import numpy as np
from . import utils


class AirFare(object):
  """This class contains the information for flights."""

  def __init__(self, fact_obj, airport_list, flight_number, ref_departure_date,
               ref_return_date):
    # 1. origin and destination
    self.origin = random.randint(0, len(airport_list) - 1)
    self.dest = random.randint(0, len(airport_list) - 1)
    if self.dest == self.origin:
      self.dest = (self.dest + 1) % len(airport_list)
    self.dest = airport_list[self.dest]
    self.origin = airport_list[self.origin]
    # 2. date and time
    d1 = np.random.normal(ref_departure_date, fact_obj.time_deviation, 1)[0]
    d2 = np.random.normal(ref_return_date, fact_obj.time_deviation, 1)[0]

    # makes ure that departure date comes first
    if d1 < d2:
      self.departure_date = d1
      self.return_date = d2
    else:
      self.departure_date = d2
      self.return_date = d1

    # assert self.departure_date <= self.return_date
    # assert self.return_date - self.departure_date <= 3600*24*365/2

    # 3. class
    self.flight_class = utils.choice(
        fact_obj.class_member, 1, p=fact_obj.class_prior)[0]

    # 4. connections
    self.connection = utils.choice(
        fact_obj.connection_member, 1, p=fact_obj.connection_prior)[0]

    # 5. price
    mean_base_price = fact_obj.flight_price_mean[self.flight_class]
    base = mean_base_price * fact_obj.low_cost_mean_fraction[self.flight_class]
    self.price = int(
        np.random.normal(
            base, base * fact_obj.flight_price_norm_std[self.connection]))
    self.price = utils.discrete_price(self.price)

    # 6. flight number
    # self.flight_number = random.randint(1000, 9999)
    self.flight_number = flight_number

    # 7. airline
    airline_ind = random.randint(0, len(fact_obj.airline_list) - 1)
    self.airline = list(fact_obj.airline_list.keys())[airline_ind]

    # post process
    self.departure_month, self.departure_day = utils.get_month_and_day(
        fact_obj, self.departure_date)
    self.return_month, self.return_day = utils.get_month_and_day(
        fact_obj, self.return_date)
    self.departure_time_num = utils.get_hour(self.departure_date)
    self.return_time_num = utils.get_hour(self.return_date)


  def get_json(self):
    return {
        'departure_airport': self.origin,
        'return_airport': self.dest,
        'departure_month': self.departure_month,
        'return_month': self.return_month,
        'departure_day': self.departure_day,
        'return_day': self.return_day,
        'departure_time_num': self.departure_time_num,
        'return_time_num': self.return_time_num,
        'class': self.flight_class,
        'num_connections': self.connection,
        'price': self.price,
        'flight_number': self.flight_number,
        'airline': self.airline
    }


class Knowledgebase(object):
  """This class contains a collection of flights."""

  def __init__(self, fact_obj, num_flights, airport_list, departure_date,
               return_date):
    self.knowledgebase = []
    base_flight_num = 1000
    for i in range(num_flights):
      self.knowledgebase.append(
          AirFare(fact_obj, airport_list, base_flight_num + i, departure_date,
                  return_date))
    has_reserv = np.random.random() < fact_obj.has_reservation_probability
    self.has_reservation = has_reserv
    if self.has_reservation:
      ind = np.random.randint(0, len(
          self.knowledgebase))  # note the numpy random above behavior
      self.reservation = self.knowledgebase[ind].flight_number
    else:
      self.reservation = 0  # 0 serves as the null number

  def get_json(self):
    knowledge_base_json = []
    for flight in self.knowledgebase:
      knowledge_base_json.append(flight.get_json())
    wrapper_json = {'kb': knowledge_base_json, 'reservation': self.reservation}
    return wrapper_json



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This is the main module that generates simulated dialogues."""

import argparse
import json
from tensorflow.compat.v1 import gfile as gfile
import tensorflow.compat.v1 as tf
import sys

from airdialogue.context_generator import context_generator_lib
from airdialogue.context_generator.src import utils

from airdialogue.simulator import interaction

FLAGS=None

def add_arguments(parser):
  """Build ArgumentParser."""
  parser.register("type", "bool", lambda v: v.lower() == "true")
  # parameters for context generator
  parser.add_argument(
      "--num_candidate_airports",
      type=int,
      default=3,
      help="number of candidate airports")
  parser.add_argument(
      "--book_window", type=int, default=2, help="window between bookings")
  parser.add_argument(
      "--num_db_record",
      type=int,
      default=30,
      help="number of database records per booking.")
  parser.add_argument(
      "--firstname_file",
      type=str,
      default="./data/resources/meta_context/first_names.txt",
      help="text file that contains a list of first names.")
  parser.add_argument(
      "--lastname_file",
      type=str,
      default="./data/resources/meta_context/last_names.txt",
      help="text file that contains a list of last names.")
  parser.add_argument(
      "--airportcode_file",
      type=str,
      default="./data/resources/meta_context/airport.txt",
      help="text file that contains a list of airport codes.")
  # parameters for simulator
  parser.add_argument("--num_samples", type=int, default=320000,
                      help="number of samples to generate.")
  parser.add_argument("--output_data", type=str, default=None,
                      help="path of the output data file.")
  parser.add_argument("--output_kb", type=str, default=None,
                      help="path of the output kb file.")
  parser.add_argument("--display_freq", type=int, default=50000,
                      help="display frequency for information.")
  parser.add_argument("--verbose", type="bool", nargs="?", const=True,
                      default=False,
                      help="if enabled, debug info will be printed out.")


def main(FLAGS):
  if FLAGS.verbose:
    print("Number of samples to generate: ", FLAGS.num_samples)
    print("Output_kb: ", FLAGS.output_kb)
    print("Output_data: ", FLAGS.output_data)

  num_samples = FLAGS.num_samples
  cg = context_generator_lib.ContextGenerator(
      num_candidate_airports=FLAGS.num_candidate_airports,
      book_window=FLAGS.book_window,
      num_db_record=FLAGS.num_db_record,
      firstname_file=FLAGS.firstname_file,
      lastname_file=FLAGS.lastname_file,
      airportcode_file=FLAGS.airportcode_file)
  inter = interaction.Interaction(
      cg.fact_obj,
      skip_greeting=0,
      fix_response_candidate=True,
      first_ask_prob=0,
      random_respond_error=True)
  ct, stats = cg.generate_context(num_samples, output_object=True)
  if FLAGS.verbose:
    print(stats)
  with gfile.Open(FLAGS.output_data,
                  "w") as f_data, gfile.Open(FLAGS.output_kb, "w") as f_kb:
    for i in range(len(ct)):
      if FLAGS.verbose and i % 5000 == 0:
        print((i, "/", len(ct)))
      cus, kb, expected_action = ct[i]
      # action has been standarlized in inter
      utterance, action, _ = inter.generate_dialogue(cus, kb)
      standarlized_intent = utils.standardize_intent(cus.get_json())
      standarlized_action = utils.standardize_action(action)
      standarlized_expected_action = utils.standardize_action(expected_action)
      # syntherized data is 100% correct. However, action contains at most one
      # flight. expected_action may contain more than one flight.
      f_data.write(
          json.dumps({"intent": standarlized_intent,"dialogue": utterance,
                      "action": standarlized_action,
                      "expected_action": standarlized_expected_action}) + "\n")
      dumped_kb = kb.get_json()
      f_kb.write(json.dumps(dumped_kb) + "\n")


def run_main(unused):
  main(FLAGS)


if __name__ == "__main__":
  this_parser = argparse.ArgumentParser()
  add_arguments(this_parser)
  FLAGS, unparsed = this_parser.parse_known_args()
  # main(FLAGS)
  tf.app.run(main=run_main, argv=[sys.argv[0]] + unparsed)






# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file only tokenize one file at a time."""

import argparse
import os
from tensorflow.compat.v1 import gfile
import tensorflow.compat.v1 as tf

from airdialogue.prepro.tokenize_lib import list_of_action_tokens_except_name
from airdialogue.prepro.tokenize_lib import flatten_json
from airdialogue.prepro.tokenize_lib import process_kb
from airdialogue.prepro.tokenize_lib import process_main_data
from airdialogue.prepro.tokenize_lib import word_tokenize
from airdialogue.prepro.tokenize_lib import write_cat
from airdialogue.prepro.tokenize_lib import write_data
from airdialogue.prepro.tokenize_lib import write_self_play
from airdialogue.prepro.tokenize_lib import write_vocabulary
# Standardization libs
from airdialogue.prepro.standardize_data_lib import standardize_and_drop
from airdialogue.prepro.standardize_data_lib import load_and_drop
import sys
import json
FLAGS = None

def add_arguments(parser):
  '''Build ArgumentParser.'''
  parser.register('type', 'bool', lambda v: v.lower() == 'true')
  parser.add_argument('--data_file', type=str, default=None,
                      help='path for data_file')
  parser.add_argument('--kb_file', type=str, default=None,
                      help='path for kb_file')
  parser.add_argument('--output_prefix', type=str, default='dev',
                      help='prefix of output file')
  parser.add_argument('--output_dir', type=str, default=None,
                      help='output dir')
  parser.add_argument('--verbose', type='bool', nargs='?', const=True,
                      default=False,
                      help='if enabled, debug info will be printed out.')
  parser.add_argument('--keep_incorrect', type='bool', nargs='?', const=True,
                      default=False,
                      help='''if enabled, incorrect dialogues will not be
                              filtered out. In stead, they will be generated
                              along with the correct ones. ''')

def json_dump(o):
  return json.dumps(o, separators = (',', ':'))

def write_infer_json(data, kb, output_file_src, output_file_tgt
  , output_file_kb):
  """This function write both kb and main data into the files."""
  f_src = gfile.Open(output_file_src, 'w')
  f_tgt = gfile.Open(output_file_tgt, 'w')
  f_kb = gfile.Open(output_file_kb, 'w')
  for entry, entry_kb in zip(data, kb):
    entire_dialogue = entry['dialogue'][:]

    # random_turn = random.randint(0, len(start) - 1)
    for target_turn in range(len(entire_dialogue))[1:]:
      f_kb.write(json_dump(entry_kb) + '\n')
      entry['dialogue'] = entire_dialogue[0:target_turn]
      f_src.write(json_dump(entry) + '\n')
      f_tgt.write(json_dump({'response': entire_dialogue[target_turn]}) + '\n')

  f_src.close()
  f_tgt.close()
  f_kb.close()

def main(FLAGS):
  output_dir = FLAGS.output_dir
  if FLAGS.verbose:
    print('output_dir', output_dir)
    print('data_file', FLAGS.data_file)
    print('kb_file', FLAGS.kb_file)
    print('output_prefix', FLAGS.output_prefix)

  if not tf.io.gfile.isdir(output_dir):
    gfile.MkDir(output_dir)

  input_data_file = FLAGS.data_file
  input_kb_file = FLAGS.kb_file
  if len(FLAGS.output_prefix.strip()) == 0:
    FLAGS.output_prefix = ''
  else:
    FLAGS.output_prefix = FLAGS.output_prefix

  output_data_pattern = output_dir + '/{0}data.json'
  output_kb_pattern = output_dir + '/{0}kb.json'

  # load data and do standardization
  raw_data, raw_kb = load_and_drop(
      input_data_file,
      input_kb_file,
      drop_incorrect=not FLAGS.keep_incorrect,
      verbose=FLAGS.verbose)

  write_infer_json(
      raw_data, raw_kb,
      output_data_pattern.format(FLAGS.output_prefix + '_infer_src_'),
      output_data_pattern.format(FLAGS.output_prefix + '_infer_tgt_'),
      output_kb_pattern.format(FLAGS.output_prefix+ '_infer_'))


def run_main(unused):
  main(FLAGS)


if __name__ == '__main__':
  this_parser = argparse.ArgumentParser()
  add_arguments(this_parser)
  FLAGS, unparsed = this_parser.parse_known_args()
  # print FLAGS
  tf.app.run(main=run_main, argv=[sys.argv[0]] + unparsed)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module builds a falsk server for visualization."""

import argparse
from os.path import expanduser
from tensorflow.compat.v1 import gfile
from collections import Counter
import nltk
import numpy as np
import json
import sys

from airdialogue.prepro.tokenize_lib import tokenize_kb

from airdialogue.evaluator.metrics.f1 import f1_score
from airdialogue.evaluator.metrics.bleu import compute_bleu
from airdialogue.evaluator.infer_utils import evaluate as evaluate_infer
from airdialogue.evaluator.selfplay_utils import compute_reward

from tqdm import tqdm
import tensorflow.compat.v1 as tf

FLAGS = None


def add_arguments(parser):
  """Build ArgumentParser."""
  parser.add_argument(
      '--true_data', type=str, default='', help='Path to the true data file.')
  parser.add_argument(
      '--true_kb', type=str, default='', help='Path to the kb file.')
  parser.add_argument(
      '--pred_data', type=str, default='', help='Path to the prediction file.')
  parser.add_argument(
      '--task',
      type=str,
      default='infer',
      help='type of the task, one of |human|infer|selfplay|')
  parser.add_argument(
      '--infer_metrics',
      type=str,
      default='bleu:brief',
      help='For infer task, choose one of multiple metric in (bleu:all|rouge:all|kl:all) or (bleu:brief|kl:brief),'
      ' this will give you a single number metric. (bleu|kl) is equivalent to (belu:brief|kl:brief) ')
  parser.add_argument(
      '--output',
      type=str,
      default='score.json',
      help='output path for score json.')

def score_human_data(flags):
  assert flags.true_data and flags.true_kb
  scores = []
  expanded_kb = expanduser(flags.true_kb)
  expanded_data = expanduser(flags.true_data)
  f2 = gfile.Open(expanded_kb)
  with gfile.Open(expanded_data) as f:
    for line in tqdm(f):
      a = json.loads(line)
      kb_line = f2.readline()
      if a['correct_sample'] == False:
        pred_action = action_obj_to_str(a['action'])
        true_action = action_obj_to_str(a['expected_action'])
        kb = tokenize_kb(json.loads(kb_line))
        ss = compute_reward(pred_action, true_action, kb)
        scores.append(ss)
      else:
        scores.append([1, 1, 1, 1])
  sn = np.array(scores)
  # np.mean(sn[:,0]), np.mean(sn[:,1]),np.mean(sn[:,2]),np.mean(sn[:,3])
  score = np.mean(sn[:, 0])
  print('final score', score)
  return {'score': score}


def score_inference(flags):
  assert flags.true_data and flags.pred_data
  expanded_true_data = expanduser(flags.true_data)
  expanded_pred_data = expanduser(flags.pred_data)

  infer_metrics = flags.infer_metrics.split(',')
  results = {}

  for metric in infer_metrics:
      infer_result = evaluate_infer(expanded_true_data, expanded_pred_data, metric)
      metric = metric.split(":")[0]
      print('infer ', metric, ': ', infer_result)
      results[metric] = infer_result
  return results

def action_obj_to_str(o):
  fl = ['empty']
  if 'flight' in o and o['flight']:
    fl = [str(f) for f in o['flight']]
  if 'name' not in o:
    o['name'] = '<unk> <unk>'
  if 'status' not in o:
    o['status'] = 'unk'
  return o['name'], "_".join(['<fl_' + f + '>' for f in fl]), '<st_' + o['status'] + '>'

def json_obj_to_tokens(o):
  d = o['dialogue']
  decapped = [' '.join(s.split(':')[1:]).strip() for s in d]
  one_string = ' '.join(decapped)
  tokenized = nltk.word_tokenize(one_string)
  return tokenized


def score_selfplay(flags):
  assert flags.true_data and flags.true_kb and flags.pred_data
  # check output

  all_score = []
  bleu_scores = []
  with tf.gfile.GFile(flags.pred_data) as f:
    with tf.gfile.GFile(flags.true_data) as t:
      with tf.gfile.GFile(flags.true_kb) as kb:
        for pred_line, true_line, kb_line in tqdm(list(zip(f, t, kb))):
          pred_json_obj = json.loads(pred_line)
          true_json_obj = json.loads(true_line)
          kb = tokenize_kb(json.loads(kb_line))
          pred_action = ''
          if 'action' not in pred_json_obj:
            pred_action = '<unk> <unk> <unk> <unk>'.split(' ')
          else:
            pred_action = action_obj_to_str(pred_json_obj['action'])
          true_action = action_obj_to_str(true_json_obj['expected_action'])
          score = compute_reward(pred_action, true_action, kb)
          all_score.append(score)

          pred_raw_text = json_obj_to_tokens(pred_json_obj)
          true_raw_text = json_obj_to_tokens(true_json_obj)

          b = compute_bleu([[true_raw_text]], [pred_raw_text])
          bleu_scores.append(b[0] * 100)

  avg_score = np.mean(all_score)
  avg_bleu = np.mean(bleu_scores)
  print('score=', avg_score)
  print('bleu=', avg_bleu)

  return {'score': avg_score, 'bleu': avg_bleu}


def main(flags):
  if flags.task == 'human':
    score = score_human_data(flags)
  elif flags.task == 'infer':
    score = score_inference(flags)
  else:
    score = score_selfplay(flags)

  with tf.gfile.GFile(flags.output, 'w') as f:
    f.write(json.dumps(score))


def run_main(unused):
  main(FLAGS)


if __name__ == '__main__':
  this_parser = argparse.ArgumentParser()
  add_arguments(this_parser)
  FLAGS, unparsed = this_parser.parse_known_args()
  tf.app.run(main=run_main, argv=[sys.argv[0]] + unparsed)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility function for the visualizer."""
import json
from tensorflow.compat.v1 import gfile


def generate_kv_nested_html(nested_kv, space):
  """This function generates html for nested key-value pair structures."""
  code = """<table style="width:{0}">\n""".format(space)
  for element in nested_kv:
    sorted_list = sorted(list(element.items()), key=lambda a: a[0])
    code = code + """
    <tr>
    """
    for el in sorted_list:
      code = code + '<td>' + str(el[0]) + ',' + str(el[1]) + '</td>\n'
    code = code + """
    </tr>
    """
  code += ' </table>\n'
  return code


def format_simple_string(kv):
  """This function format a simple key-varlue pair.
  It also highlight some of the fields in intent.
  """
  sorted_list = sorted(list(kv.items()), key=lambda a: a[0])
  arr = []
  for el in sorted_list:
    key = str(el[0])
    if key in ['return_time', 'departure_time']:
      key = """<font style="color: #fff; background-color: red">""" + key + """</font>"""
    if key == 'airline_preference':
      key = """<font style="color: #fff; background-color: blue">""" + key + """</font>"""
    arr.append((key + ':' + str(el[1])))
  return ', '.join(arr)


def generate_form(partitions, sid, partition):
  """This function generate the submit form."""

  sel_html = """
  Partition <select name="partition">\n"""
  for p in partitions:
    if p == partition:
      selected = """ selected="selected" """
    else:
      selected = ''
    sel_html += """<option {0} value="{1}">{1}</option>\n""".format(selected, p)
  sel_html += '</select>\n'

  ret = """<form method="POST" action="/">""" + sel_html + '<br>'
  ret += """Sample #<input type="text" name="index" value="{0}" required><br>
            <input type="submit" value="Submit">
            </form>""".format(sid)
  return ret


def generate_res_str(reservation):
  if reservation == 0:
    return 'This customer do not have an existing reservation.'
  else:
    return 'This customer has a reservation on flight #{0}'.format(reservation)


def generate_html(sample,
                  kb_object,
                  sid,
                  partitions,
                  partition):
  """This function generates the html."""
  intent, action, true_action = sample['intent'], sample['action'], sample[
      'expected_action']
  if 'dialogue' in sample:
    dialogue = sample['dialogue']
  else:
    dialogue = None
  db = kb_object['kb']
  reservation = kb_object['reservation']
  res_str = generate_res_str(reservation)
  # print ('reservation', reservation)
  if 'search_info' in sample:
    search_info = sample['search_info']
    if type(search_info) == str:
      search_info = json.loads(search_info)
  else:
    search_info = None
  # print('intent', intent.keys())
  intent_table = format_simple_string(intent)
  action_table = format_simple_string(action)
  true_action_table = format_simple_string(true_action)

  if search_info:
    # print('search_info', search_info)
    search_info_table = generate_kv_nested_html(search_info, '100%')
  else:
    search_info_table = None

  db_table = generate_kv_nested_html(db, '100%')
  html_code = """
  <!DOCTYPE html>
  <html>
  <body>
  AirDialogue Dataset Visualizer"""
  html_code += generate_form(partitions, sid, partition)
  html_code += """<h2>AirDialogue No. {0} on partition {1}. </h2>\n
  """.format(sid, partition)
  html_code = str(html_code)
  html_code += """<table style="width:100%">"""
  html_code += """<tr> {0}</tr>""".format(
      '<strong>intent:</strong>' + intent_table.encode('ascii', 'ignore') +
      '<br>')
  html_code += """<tr> {0}</tr>""".format(
      '<strong>user action:</strong>' + action_table.encode('ascii', 'ignore') +
      '<br>')
  html_code += """<tr> {0}</tr>""".format(
      '<strong>true action:</strong>' +
      true_action_table.encode('ascii', 'ignore') + '<br>')
  html_code += '<strong>' + res_str + '</strong><br>'
  html_code += """<tr> {0}</tr>""".format('<strong>db table:</strong>' +
                                          db_table.encode('ascii', 'ignore') +
                                          '<br>')

  if search_info_table:
    html_code += """<tr> {0}</tr>""".format(
        '<strong>search info:</strong>' +
        search_info_table.encode('ascii', 'ignore') + '<br>')

  if dialogue:
    html_code += """<tr> {0}</tr>""".format('<p>'.join(dialogue).encode(
        'ascii', 'ignore'))

  html_code += """</table>"""
  html_code += """
  </body>
  </html>"""
  return html_code


def make_path(path):
  if not gfile.Exists(path):
    gfile.MkDir(path)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains the main dialogue logic."""

import numpy as np
from airdialogue.context_generator.src import utils
greeting_pairs = [('Hello.', 'Hi.'), ('Hey, how are you.',
                                      'I am fine. Thanks for asking.'),
                  ('Hello there.', 'Hey!'), ('Hey there.', 'Hi.'),
                  ('Hi.', 'Hello.'), ('How is it going?',
                                      'I am good. Thank you.')]

# spaces will be processed in tokenizer
agent_ask = ['How can I help you today?', 'What can I do for you today?']

customer_request = {
    'book': [
        'I would like to book a flight {0}.', 'I want to book a flight {0}.',
        'Please help me to book a flight {0}.',
        'Could you help me to book a flight {0}?'
    ],
    'cancel': [
        'Could you help me to cancel my flight?',
        'Please help me to cnacel my reservation.',
        'I would like to cancel my reservation.'
    ],
    'change': [
        'I would like to change my reservation.',
        'Could you help me to change my recent flight reservation?',
        'I need helps to change my reservation.'
    ]
}

# c1=departure city
# d1= departure date
# c2=return city
# d2=return date
agent_first_respond = {
    'book': {
        'c1c2': [
            'Sure, may I start with the departure and arrival city',
            'Not a problem. What\'s your departure and arrival city?',
            'Happy to do so. Could you tell me your departure and arrival city?'
        ],
        'd1d2': [
            'What\'s the departure and return date?',
            'What day do you want to depart and return?'
        ]
    },
    'change': {
        'c1c2': [
            'Sure. What\' the departure and arrival city for your updated '
            'reservation?',
            'Not a problem. Could you tell me the departure and arrival city '
            'for your updated reservation?',
            'I see. Could you tell me the departure and arrival city of your '
            'updated reservation?'
        ],
        'd1d2': [
            'What\'s the departure and return date?',
            'What day do you want to depart and return?'
        ]
    }
}

customer_first_respond = {
    'c1c2': [
        'My departure city is {0} and my return city is {1}.',
        'I will be flying from {0} to {1}.',
        'Departure city is {0} and return city is {1}.'
    ],
    'd1d2': [
        'I will be departing on {0} and returning on {1}.',
        'I am plnning to depart on {0} and return on {1}.'
    ]
}

# agent_recurrening_respond={['What\'s your {0}?',
#                            'Could you let me know your {0}?',
#                            'I will need to know your {0}.']}
customer_recurring_respond = {
    'departure_time': [
        'I am actually considering departing in the {0}.',
        'Is there any flight that departs in the {0}?',
        'I am actually thinknig about departing in the {0}.'
    ],
    'return_time': [
        'I would prefer to return in the {0}.',
        'Can you find a flight that returns in the {0}?'
    ],
    'class': [
        'I am looking for a {0} class flight.',
        'Do you have any {0} class flights available?'
    ],
    'max_connections': [
        'I prefer no more than {0} connections.',
        'Could you find me flights that are less than or equal to {0} '
        'connections?'
    ],
    'max_price': [
        'I have a strict budget constraint. Could you find me flights that are'
        ' below ${0}.',
        'My budget limits me to flights that are cheaper than ${0}.'
    ],
    'airline_preference': [
        'Could you help me to find a normal fare airline other than a low-cost'
        ' one?', 'I prefer not to fly low-cost airlines.'
    ],
    'satisfied': ['This flight looks good.', 'Sounds good.', 'Looking good.']
}

agent_suggestion = {
    'full': [
        'We have flight {0} departs from {1} on {2} at {3} and returns from '
        '{4} on {5} at {6}. This is a {7} class flight with {8}. The total '
        'price is ${9}. '
    ],
    'without_cd': [
        'I have flight {0} departs the same day at {1} and returns the same '
        'day at {2}. This is a {3} class flight with {4} and a total price of '
        '${5}.',
        'How about flight {0} departs the same day at {1} and rturns the same '
        'day at {3}. It is a {3} class with {4} and a price of ${5}.'
    ]
}

agent_ask_name = [
    'May I have the name for the reservation?',
    'What\'s the name for this reservation?',
    'Could you let me know the name for this reservation?'
]
cutomer_name = ['My name is {0}.', '{0}']

agent_confirm_change = {
    'cancel': [
        'I am able to locate your ticket. Could you confirm that you want to '
        'cancel it?',
        'I found your reservation. Are you sure you want to cancel it?',
        'Just want to confirm that you want to cancel the ticket, correct?'
    ],
    'change': [
        'I just want to confirm everything we have worked on so far. You are '
        'about to change the reservation to {0}. Could you confirm?',
        'Just one more step and I will be done. Could you confirm your change '
        'to {0}?'
    ],
    'book': [
        'We have your ticket of {0}. Do I have your permission to proceed?',
        'I just want to confirm with your on flight {0}. Is this the ticket we'
        ' are going to book?'
    ]
}

customer_confirm_change = {
    'book': [
        'Yes, please go ahead and book the ticket.',
        'Confirmed. Please go ahead.'
    ],
    'change': [
        'That is correct. Please make the cahnges.',
        'Sounds good. Please proceed'
    ],
    'abort': [
        'Actually, I have changed my mind.',
        'Well, I made a second through and I have decided to keep my original '
        'reservation.'
    ],
    'cancel': ['Yes, please cancel it.', 'Yup, please cancel it.']
}

agent_conclusion_message = {
    'cancel': [
        'Your ticket has been canceled. Thank you for using our flight booking'
        ' service.',
        'Not a problem. I have canceled your ticket. Have a good day.',
        'Your ticket is canceled.'
        ' Please let us know if you have other requests.'
    ],
    'book': [
        'Your booking has been confirmed. Thank you.',
        'Your ticket has been booked. Have a nice day.',
        'We have successfully booked your trip. Thanks for using our flight '
        'booking servie.'
    ],
    'change': [
        'I have updated your reservation. Thanks for using our flight booking '
        'service.',
        'Your reservation has been updated. Please let us know if you have '
        'other requests.',
        'We have successfully updated your reservation. Have a nice day.'
    ],
    'no_flight': [
        'I can not find any flights that satisfy your requests.',
        'There is no flight that satisfies your requests.'
    ],
    'abort': [
        'You have chosen not to proceed. Thanks for using our flight booking '
        'service.', 'Not a problem. We will keep your'
        ' reservation unchanged. Have a nice day.'
    ],
    'no_res': [
        'I am sorry, but I can not locate your reservation.',
        'Your reservation can not be found in our system. Please check your '
        'account.'
    ]
}

customer_turn_prefix = 'customer:'
agent_turn_prefix = 'agent:'
secondary_error = True


class Interaction(object):
  """This class contains the dialogue interaction between the two agents."""

  def __init__(self,
               fact_obj,
               skip_greeting=0,
               fix_response_candidate=True,
               first_ask_prob=0,
               regret_prob=0,
               random_respond_error=False):
    # probability then customer will skip always greeting if he/she speaks first
    self.skip_greeting = skip_greeting
    # if true then only the first response candidate will be used.
    # otherwise randomly decided.
    self.fix_resp_candidate = fix_response_candidate
    # probability that customer ask turn will not provide information.
    # otherwise randomly decided.
    self.first_ask_prob = first_ask_prob
    # probability of regetting making changes or purchases at the end
    self.regret_prob = regret_prob
    # randomly choose unsatisfied condition to respond
    self.random_respond_error = random_respond_error
    # fact object
    self.fact_obj = fact_obj

  def get_template(self, list_of_templates):
    if self.fix_resp_candidate:
      choice = 0
    else:
      choice = utils.choice(len(list_of_templates))
    return list_of_templates[choice]

  def customer_turn(self, utterance):
    return customer_turn_prefix + ' ' + utterance

  def agent_turn(self, utterance):
    return agent_turn_prefix + ' ' + utterance

  def generate_confirmation(self, flight, full=True):
    if full:
      tmp = self.get_template(agent_suggestion['full'])
      return tmp.format(
          flight['flight_number'], flight['departure_airport'],
          flight['departure_month'] + ' ' + str(flight['departure_day']),
          utils.format_time(
              flight['departure_time_num']), flight['return_airport'],
          flight['return_month'] + ' ' + str(flight['return_day']),
          utils.format_time(flight['return_time_num']), flight['class'],
          utils.get_connection(flight['num_connections']), flight['price'])
    else:
      tmp = self.get_template(agent_suggestion['without_cd'])
      return tmp.format(
          flight['flight_number'],
          utils.format_time(flight['departure_time_num']),
          utils.format_time(flight['return_time_num']), flight['class'],
          utils.get_connection(flight['num_connections']), flight['price'])

  # this is wrong because we will need to compare all flights. use the one in
  # context generator
  def airflight_selector(self, condition, airfare_database):
    """This function selects a flight based on the condition."""
    candidates = []
    for i, flight in enumerate(airfare_database):
      cond = utils.check_condition(
          self.fact_obj, flight, condition, get_full_diff=False)
      if cond == 'satisfied':
        candidates.append(i)
    candidates.sort(key=lambda a: airfare_database[a]['price'])
    if not candidates:
      return None
    else:
      return airfare_database[candidates[0]]

  def get_flight_from_flight_number(self, flight_number, airfare_database):
    for i, flight in enumerate(airfare_database):
      if flight['flight_number'] == flight_number:
        return airfare_database[i]
    return None

  def get_message(self, ag_condition, cus_condition, erro_state):
    """This function geenrates the message based on errors."""
    if erro_state == 'satisfied':
      error = erro_state
    else:
      if self.random_respond_error:
        error_to_handle = utils.choice(list(range(len(erro_state))))
      else:
        error_to_handle = 0
      error = erro_state[error_to_handle]
    msg = self.get_template(customer_recurring_respond[error])
    if error != 'satisfied':
      ag_condition[error] = cus_condition[error]

    if error not in set(['airline_preference', 'satisfied']):
      return msg.format(cus_condition[error]), ag_condition, error
    else:
      return msg, ag_condition, error

  def generate_greetings(self, utterance, speaker):
    """This function generates greets."""
    # skip greeting
    if speaker == 0 and np.random.random() < self.skip_greeting:
      return utterance
    pair = self.get_template(greeting_pairs)
    if speaker == 0:
      utterance.append(self.customer_turn(pair[0]))
      utterance.append(self.agent_turn(pair[1]))
    else:
      utterance.append(self.agent_turn(pair[0]))
      utterance.append(self.customer_turn(pair[1]))
    return utterance

  def generate_agent_ask(self, utterance):
    utterance.append(self.agent_turn(self.get_template(agent_ask)))
    return utterance

  def generate_customer_request(self, customer_condition, agent_condition,
                                utterance):
    """This function generates customer requests."""
    # first get the goal index (buy, cancel, change)
    goal_index = customer_condition['goal']
    goal_str = self.fact_obj.goal_str_arr[goal_index]
    # get the template
    template = self.get_template(customer_request[goal_str])
    # by some probability customer will say departure and return
    # city in the request statement
    first_ask = np.random.random() < self.first_ask_prob
    if first_ask:
      deprture_city, return_city = customer_condition[
          'departure_airport'], customer_condition['return_airport']
      ask = 'from ' + deprture_city + ' to ' + return_city
      # set agent_condition
      agent_condition['departure_airport'] = deprture_city
      agent_condition['return_airport'] = return_city
    else:
      # otherwise utt will just be empty and generaet
      # a request without filling information
      ask = ''
    utt = template.format(ask)
    utterance.append(self.customer_turn(utt))
    agent_condition['goal'] = customer_condition['goal']
    return utterance, agent_condition, goal_str

  def no_reservation(self, utterance):
    agent_confirm_utt = self.get_template(agent_conclusion_message['no_res'])
    utterance.append(self.agent_turn(agent_confirm_utt))
    return utterance

  def cancel_reservation(self, utterance):
    agent_utt = self.get_template(agent_confirm_change['cancel'])
    utterance.append(self.agent_turn(agent_utt))
    customer_utt = self.get_template(customer_confirm_change['cancel'])
    utterance.append(self.customer_turn(customer_utt))
    agent_confirm_utt = self.get_template(agent_conclusion_message['cancel'])
    utterance.append(self.agent_turn(agent_confirm_utt))
    return utterance

  def fulfill_basic_requirement(self, goal, cus_cond, ag_cond, utterance):
    """This function fulfill bsic requirements."""
    # check departure and return city
    if 'departure_airport' not in ag_cond:
      assert 'return_airport' not in ag_cond
      # print ('goal',goal)
      ask_cities = self.get_template(agent_first_respond[goal]['c1c2'])
      utterance.append(self.agent_turn(ask_cities))
      depart = cus_cond['departure_airport']
      ret = cus_cond['return_airport']
      respond_cities = self.get_template(customer_first_respond['c1c2']).format(
          depart, ret)
      utterance.append(self.customer_turn(respond_cities))
      ag_cond['departure_airport'] = depart
      ag_cond['return_airport'] = ret
    if 'departure_month' not in ag_cond:
      assert 'departure_day' not in ag_cond
      assert 'return_month' not in ag_cond
      assert 'return_day' not in ag_cond
      ask_date = self.get_template(agent_first_respond[goal]['d1d2'])
      utterance.append(self.agent_turn(ask_date))
      d_m, d_d = cus_cond['departure_month'], cus_cond['departure_day']
      a_m, a_d = cus_cond['return_month'], cus_cond['return_day']
      dep = d_m + ' ' + str(d_d)
      ar = a_m + ' ' + str(a_d)
      respond_date = self.get_template(customer_first_respond['d1d2']).format(
          dep, ar)
      utterance.append(self.customer_turn(respond_date))
      ag_cond['departure_month'] = d_m
      ag_cond['departure_day'] = d_d
      ag_cond['return_month'] = a_m
      ag_cond['return_day'] = a_d
    return ag_cond, utterance

  def continue_booking(self, cus_cond, ag_cond, kb, utterance):
    """This function books the flight."""
    goal = ag_cond['goal']
    goal_str = self.fact_obj.goal_str_arr[goal]
    ag_cond, utterance = self.fulfill_basic_requirement(goal_str, cus_cond,
                                                        ag_cond, utterance)
    status = None
    first_time = True
    error = 'basic'
    while not status:
      flight = self.airflight_selector(ag_cond, kb)
      if not flight:  # terminal condition
        utterance.append(
            self.agent_turn(
                self.get_template(agent_conclusion_message['no_flight'])))
        if secondary_error:
          status = 'no_flight' '_' + error
        else:
          status = 'no_flight'

        return utterance, status, flight
      else:
        utterance.append(
            self.agent_turn(self.generate_confirmation(flight, first_time)))
        first_time = False
        condition = utils.check_condition(self.fact_obj, flight, cus_cond)
        msg, ag_cond, error = self.get_message(
            ag_cond, cus_cond, condition)  # will do merge condition within
        utterance.append(self.customer_turn(msg))
        if condition == 'satisfied':
          goal_int = cus_cond['goal']
          status = self.fact_obj.goal_str_arr[goal_int]  # either book or change
    # status is either book, change, or potenitally abort
    ask_confirm = self.get_template(agent_confirm_change[status]).format(
        flight['flight_number'])
    utterance.append(self.agent_turn(ask_confirm))
    regret = np.random.random() < self.regret_prob
    if regret:
      status = 'abort'
    cus_re_conrim = self.get_template(customer_confirm_change[status])
    utterance.append(self.customer_turn(cus_re_conrim))
    agent_conclusion = self.get_template(agent_conclusion_message[status])
    utterance.append(self.agent_turn(agent_conclusion))
    return utterance, status, flight

  def generate_dialogue(self, customer, knowledge_base):
    """This function is the main entry of the dialogue generation logic."""
    airfare_database = knowledge_base.get_json()['kb']
    reservation = knowledge_base.get_json()['reservation']
    utterance = []
    # 0a. decides who speaks first 0--customer, 1--agent
    speaker = int(np.random.random() < 0.5)
    # 0b. generate customer's full condition and agent_condition
    customer_condition = customer.get_customer_condition()
    agent_condition = {}
    # 1. greetings
    utterance = self.generate_greetings(utterance, speaker)
    # 2. generate agent's utterance to ask for request if
    # customer finished the last turn
    if speaker == 1:
      utterance = self.generate_agent_ask(utterance)
    # 3. generate customer request
    utterance, agent_condition, goal_str = self.generate_customer_request(
        customer_condition, agent_condition, utterance)

    # 4 ask for name first
    ask_name_utt = self.get_template(agent_ask_name)
    utterance.append(self.agent_turn(ask_name_utt))
    answer_name_utt = self.get_template(cutomer_name).format(
        customer_condition['name'].replace('_', ' '))
    utterance.append(self.customer_turn(answer_name_utt))

    # 5. fulfill basic requirement
    # (departure/return city, departure/return month/day)

    if goal_str == 'book' or (goal_str == 'change' and reservation != 0):
      # status can be book, change, no_flight, abort
      utterance, status, flight = self.continue_booking(
          customer_condition, agent_condition, airfare_database, utterance)
    elif goal_str in ['change', 'cancel'] and reservation == 0:
      utterance = self.no_reservation(utterance)
      status = 'no_reservation'
      flight = None
    elif goal_str == 'cancel':
      utterance = self.cancel_reservation(utterance)
      status = 'cancel'
      flight = None

    # per new format we will need to return a flight arr to consider the
    # situation where multiple cheapest flights are available. As in the real
    # data here we only choose at most one flight. However, the
    # expected_action should be able to contain multiple flights.
    flight_arr = []
    if flight:
      flight_arr.append(flight)
    # use the generate action function from utils. It will standarlize
    # the action
    # print(customer_condition['name'])
    name = customer_condition['name'].replace('_', ' ')
    action = utils.generate_action(flight_arr, name, status)

    # print utterance
    return utterance, action, status

  # def generate_action(self, flight, name, status):
  #   action_json = {}
  #   if status.startswith('no_flight'):
  #     status = 'no_flight'
  #   action_json['status'] = status
  #   if status in ['change', 'book']:
  #     action_json['flight'] = flight['flight_number']
  #     action_json['name'] = name
  #   return action_json



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This function calculates the F1 score."""
from collections import Counter
import re
import string


def f1_score(prediction, ground_truth):
  prediction_tokens = list(normalize_answer(prediction))
  ground_truth_tokens = list(normalize_answer(ground_truth))
  common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
  num_same = sum(common.values())
  # print num_same, common
  if num_same == 0:
    return 0
  precision = 1.0 * num_same / len(prediction_tokens)
  recall = 1.0 * num_same / len(ground_truth_tokens)
  f1 = (2 * precision * recall) / (precision + recall)
  return f1


def normalize_answer(s):
  """Lower text and remove punctuation, articles and extra whitespace."""

  def remove_articles(text):
    return re.sub(r'\b(a|an|the)\b', ' ', text)

  def white_space_fix(text):
    return ' '.join(text.split())

  def remove_punc(text):
    exclude = set(string.punctuation)
    return ''.join(ch for ch in text if ch not in exclude)

  def lower(text):
    return text.lower()

  return white_space_fix(remove_articles(remove_punc(lower(s))))




# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""library file for tokenize."""

import nltk
import numpy as np
from tensorflow.compat.v1 import gfile
from tqdm import tqdm

start_of_turn1 = '<t1>'
start_of_turn2 = '<t2>'
end_of_dialogue = '<eod>'
unk_token = '<unk>'
list_of_action_tokens_except_name = set([])


def tokenize_kb(kb_json):
  """This function tokenizes the knowledge base json."""

  def tokenize_one_flight(flight_json):
    """this function tokenize one flight entry in the knowledge base."""
    a1, a2 = flight_json['departure_airport'], flight_json['return_airport']
    m1, m2 = flight_json['departure_month'], flight_json['return_month']
    d1, d2 = flight_json['departure_day'], flight_json['return_day']
    tn1, tn2 = flight_json['departure_time_num'], flight_json['return_time_num']
    cl = flight_json['class']
    pr = flight_json['price']
    cn = flight_json['num_connections']
    al = flight_json['airline']
    fl = flight_json['flight_number']
    # add tags
    a1, a2, m1, m2 = format_tag('a1', a1), format_tag('a2', a2), format_tag(
        'm1', m1), format_tag('m2', m2)
    d1, d2, tn1, tn2 = format_tag('d1', d1), format_tag('d2', d2), format_tag(
        'tn1', tn1), format_tag('tn2', tn2)
    cl = format_tag('cl', cl)
    pr = format_tag('pr', pr)
    cn = format_tag('cn', cn)
    al = format_tag('al', al)
    fl = format_tag('fl', fl)
    # combine
    arr = [a1, a2, m1, m2, d1, d2, tn1, tn2, cl, pr, cn, al, fl]
    for element in arr:
      list_of_action_tokens_except_name.add(element)
    return ' '.join(arr)

  if kb_json['reservation'] == 0:
    res = 'no_res'
  else:
    res = 'has_res'
  res = format_tag('res', res)
  flight_arr = []
  for flight in kb_json['kb']:
    tokenized_flight = tokenize_one_flight(flight)
    flight_arr.append(tokenized_flight)
  list_of_action_tokens_except_name.add(res)
  return res + ' ' + ' '.join(flight_arr)


def process_kb(raw_kb, word_map, stream=False):
  """main entry to process kb."""
  processed_data = []
  d = raw_kb
  if not stream:
    d = tqdm(raw_kb, desc='process kb')
  for kb_object in d:
    # the database will be flattened into a single sequence of tokens.
    flattened = tokenize_kb(kb_object)
    processed_data.append(flattened)
    word_map = apply_word_map(flattened, word_map)
  return processed_data, word_map


def flatten_json(obj):
  if isinstance(obj, dict):
    new_list = sorted(list(obj.items()), key=lambda a: a[0])
    res_list = []
    for key, value in new_list:
      new_val = flatten_json(value)
      res_list.append(key + ' : ' + str(new_val))
    return ' . '.join(res_list)
  if isinstance(obj, list):
    new_list = []
    for x in obj:
      new_list.append(flatten_json(x))
    return ' . '.join(new_list)
    # return sorted(order_json(x) for x in obj)
  else:
    return obj


def apply_word_map(flattened_sent, word_map):
  words = flattened_sent.split(' ')
  for w in words:
    if w not in word_map:
      word_map[w] = 1
    else:
      word_map[w] += 1
  return word_map


def format_tag(name, val):
  return '<{0}_{1}>'.format(str(name), str(val))


def get_full_intent(intent_json):
  """recovers the full intent json from standalized intent.

  Basically we will add fields that are omitted becauase their values are
  all or 2 back to the intent json.
  """
  # dep/ret time
  if 'departure_time' not in intent_json:
    intent_json['departure_time'] = 'all'
  if 'return_time' not in intent_json:
    intent_json['return_time'] = 'all'
  # class
  if 'class' not in intent_json:
    intent_json['class'] = 'all'
  # max_connections
  if 'max_connections' not in intent_json:
    intent_json['max_connections'] = 2
  # airline_preference
  if 'airline_preference' not in intent_json:
    intent_json['airline_preference'] = 'all'
  return intent_json


def add_dash_to_name(original_name):
  return '_'.join(original_name.split(' '))


def tokenize_intent(intent_json):
  """This function tokenizes intents."""
  a1, a2 = intent_json['departure_airport'], intent_json['return_airport']
  m1, m2 = intent_json['departure_month'], intent_json['return_month']
  d1, d2 = intent_json['departure_day'], intent_json['return_day']
  t1, t2 = intent_json['departure_time'], intent_json['return_time']
  # nm = add_dash_to_name(intent_json['name'])
  nm1, nm2 = intent_json['name'].strip().split(' ')
  cl = intent_json['class']
  pr = intent_json['max_price']
  cn = intent_json['max_connections']
  al = intent_json['airline_preference']
  gl = intent_json['goal']
  # add tags
  a1, a2, m1, m2 = format_tag('a1', a1), format_tag('a2', a2), format_tag(
      'm1', m1), format_tag('m2', m2)
  d1, d2, t1, t2 = format_tag('d1', d1), format_tag('d2', d2), format_tag(
      't1', t1), format_tag('t2', t2)
  # nm=format_tag('nm',nm)
  cl = format_tag('cl', cl)
  pr = format_tag('pr', pr)
  cn = format_tag('cn', cn)
  al = format_tag('al', al)
  gl = format_tag('gl', gl)
  # concatenates
  arr = [a1, a2, m1, m2, d1, d2, t1, t2, nm1, nm2, cl, pr, cn, al, gl]
  return ' '.join(arr)


def tokenize_action(action_json, first_name_cat, last_name_cat, flight_cat,
                    state_cat):
  """Both name and flight will always be in the action.

  Context might have
  multiple flights in the arr but training/eval/testing data should only have
  one arr in the flight arr.
  """
  try:
    name_arr = action_json['name'].strip().split(' ')
    if len(name_arr) < 2:  # name_arr has at least one element
      nm1 = name_arr[0]
      nm2 = ''
    else:
      nm1, nm2 = name_arr
  except:
    print('name', action_json['name'])
  fl_arr = action_json['flight']
  if len(fl_arr) == 0:
    fl = 'empty'
    fl = format_tag('fl', fl)
  else:
    fl_arr_str = []
    for f in fl_arr:
      flight = format_tag('fl', str(f))
      fl_arr_str.append(flight)
      flight_cat.add(str(flight).strip())
      # fl_arr_str.append(str(f).strip())
    fl = '_'.join(fl_arr_str)
  st = action_json['status']
  # add tags
  # fl = format_tag('fl', fl)
  st = format_tag('st', st)
  arr = [nm1, nm2, fl, st]
  # list_of_action_tokens.add(nm)
  list_of_action_tokens_except_name.add(fl)
  list_of_action_tokens_except_name.add(st)
  first_name_cat.add(nm1.strip())
  last_name_cat.add(nm2.strip())
  state_cat.add(st)
  return ' '.join(arr)


# Right now expected action is not used only one flight is considered.
def process_main_data(raw_data,
                      sent_tok,
                      word_tok,
                      word_map,
                      input_type,
                      stream=False,
                      self_play_start_turn=None):
  """This function processes the main data."""

  def process_dialogue(dialogue):
    """This function processes dialogues."""

    def do_tokenize(last, turn):
      """This function tokenizes the diloagues of a specific turn."""
      if turn.startswith('customer: '):
        # agent:
        turn = turn[10:].strip()
        sot = start_of_turn1
        eot = start_of_turn2
      else:  # agent:
        turn = turn[7:].strip()
        sot = start_of_turn2
        eot = start_of_turn1
      sentences = sent_tok(turn)
      tokenized_sents = []
      for s in sentences:
        words = word_tok(s)
        tokenized_sents.append(' '.join(words))
      flat_content = ' '.join(tokenized_sents)
      if last:
        return sot + ' ' + flat_content + ' ' + end_of_dialogue + ' ' + eot
      else:
        return sot + ' ' + flat_content

    tokenized_dialogue = []
    max_turn_length = 0
    for i, turn in enumerate(dialogue):
      tokenized_turn = do_tokenize(i == len(dialogue) - 1, turn)
      max_turn_length = max(max_turn_length, len(tokenized_turn.split(' ')))
      tokenized_dialogue.append(tokenized_turn)

    return ' '.join(tokenized_dialogue), max_turn_length

  def get_dialogue_boundary(start_token, flat_dialogue):
    """This function gets the boundary array of the dialogues."""

    def get_end_token(start, set_of_end_tokens, splitted_dialogues):
      for i in range(start, len(splitted_dialogues)):
        if splitted_dialogues[i] in set_of_end_tokens:
          return i
      assert False, 'end token not found : ' + ' '.join(
          flat_dialogue) + 'start=' + str(start) + '/' + str(
              len(splitted_dialogues))

    def get_next_start_token(end_position, start_token, splitted_dialogues):
      for i in range(end_position, len(splitted_dialogues)):
        if splitted_dialogues[i] == start_token:
          return i
      return len(splitted_dialogues)

    set_of_end_tokens = set([
        start_of_turn1, start_of_turn2
    ])  # taking out end_of_dialogue token because of dynamic rnn decoder
    # set_of_end_tokens=set([start_of_turn1,start_of_turn2,end_of_dialogue])
    splitted_dialogue = flat_dialogue.split(' ')
    i = get_next_start_token(0, start_token, splitted_dialogue)
    all_starts = []
    all_ends = []
    while i < len(splitted_dialogue
                 ) - 1:  # we don't find the end token for the last turn change.
      end_position = get_end_token(i + 1, set_of_end_tokens, splitted_dialogue)
      err_msg = 'start token appeared twice: ' + flat_dialogue
      assert splitted_dialogue[end_position] != start_token, err_msg
      # all_res.append((i,end_position))
      all_starts.append(i)
      all_ends.append(end_position)
      i = get_next_start_token(i + 1, start_token, splitted_dialogue)
    return (all_starts, all_ends)

  def serialize_boundary(start, end):
    all_res = []
    for s in start:
      all_res.append(str(s))
    for e in end:
      all_res.append(str(e))
    return ' '.join(all_res)

  intents = []
  actions = []
  expected_actions = []
  dialogues = []
  boundaries1 = []
  boundaries2 = []
  lengths = []
  max_diag_length = 0
  max_turn1 = 0
  first_name_cat = set([])
  last_name_cat = set([])
  flight_cat = set([])
  state_cat = set([])

  d = raw_data
  if not stream:
    d = tqdm(raw_data, desc='process raw data')

  for loaded_json in d:
    # loaded_json = json.loads(delete_non_ascii(line))
    # input_type
    if 'intent' in loaded_json:
      intent = loaded_json['intent']
      processed_intent = tokenize_intent(get_full_intent(intent))
      intents.append(processed_intent)
      word_map = apply_word_map(processed_intent, word_map)
    else:
      intents.append('')
    if 'action' in loaded_json and loaded_json['action']:
      action = loaded_json['action']
      processed_action = tokenize_action(
          action,
          first_name_cat,
          last_name_cat,
          flight_cat,
          state_cat,
      )
      actions.append(processed_action)
      word_map = apply_word_map(processed_action, word_map)
    else:
      actions.append('')
    if 'expected_action' in loaded_json:
      expected_action = loaded_json['expected_action']
      # print "processed_intent", processed_intent
      processed_expected_action = tokenize_action(
          expected_action,
          first_name_cat,
          last_name_cat,
          flight_cat,
          state_cat,
      )
      # print "processed_action", processed_action
      expected_actions.append(processed_expected_action)

    # NB for word map updates above:
    # word map should contain everything in the supervised eval and training
    # however, it should not contain expected_action since this is for self-play
    # it contains tokens that has _, on flights.

    # process dialogue only when input_type is dialogue
    if input_type == 'dialogue':
      dialogue = loaded_json['dialogue']
      processed_dialogue, max_diag_len_this = process_dialogue(dialogue)
      max_diag_length = max(max_diag_length, max_diag_len_this)
      t1_boundaries = get_dialogue_boundary(start_of_turn1, processed_dialogue)
      t2_boundaries = get_dialogue_boundary(start_of_turn2, processed_dialogue)
      max_turn1 = max(max_turn1, len(t1_boundaries[0]) + len(t2_boundaries[0]))
      boundaries1.append(serialize_boundary(*t1_boundaries))
      boundaries2.append(serialize_boundary(
          *t2_boundaries))  # this is used only for inference files generation
      if processed_dialogue == '' and \
          self_play_start_turn in ['agent', 'customer']:
        if self_play_start_turn == 'agent':
          processed_dialogue = '<t2>'
        else:
          processed_dialogue = '<t1>'

      dialogues.append(processed_dialogue)
      length = len(processed_dialogue.split(' '))
      lengths.append(length)
      word_map = apply_word_map(processed_dialogue, word_map)

  if input_type == 'dialogue' and not stream:  #  output stats only when input is dialogue
    min_length, mean_length, max_length = np.min(lengths), np.mean(
        lengths), np.max(lengths)
    print(
        ('min_len: ${0}, mean_len: {1}, max_len: {2}'
         'max_sent_len: {3}, max_turn: {4}').format(min_length, mean_length,
                                                    max_length, max_diag_length,
                                                    max_turn1))

  # return all the processed data. Some of them will be empty arrays when in
  # context mode.
  all_cat = [first_name_cat, last_name_cat, flight_cat, state_cat]
  return intents, actions, expected_actions, dialogues, word_map, boundaries1, boundaries2, all_cat


def is_ascii(s, keep_non_ascii):
  if keep_non_ascii:
    return True
  else:
    return all(ord(c) < 128 for c in s)


def write_vocabulary(output_file, output_all_vocab_file, word_frequency,
                     frequency_cutoff, keep_non_ascii):
  special_chars = [unk_token, start_of_turn1, start_of_turn2, end_of_dialogue]
  new_word_frequency = set([])
  # if output_filei not None, we write to file, otherwise we don't.
  if output_file:
    f = gfile.Open(output_file, 'w')
  else:
    f = None

  # first one should always be unktoken
  for special_char in special_chars:
    if f:
      f.write(special_char + '\n')
  for key in word_frequency:
    # We write to the vocabulary only when the key is not empty.
    # Otherwise tensorflow will complain.
    if word_frequency[key] >= frequency_cutoff and (
        key not in special_chars) and is_ascii(key, keep_non_ascii) and key:
      if f:
        f.write(key + '\n')
      new_word_frequency.add(key)

  if f:
    f.close()
  # all vocab
  if output_all_vocab_file:
    with gfile.Open(output_all_vocab_file, 'w') as f2:
      f2.write(str(word_frequency))
  return new_word_frequency


def write_cat(files, cats):
  for file, category in zip(files, cats):
    with gfile.Open(file, 'w') as f:
      for cat in category:
        f.write(str(cat) + '\n')


def write_data(data, output_file_data, output_file_kb, alt_infer=False):
  """This function writes data into a text file."""
  f_data = gfile.Open(output_file_data, 'w')
  f_kb = gfile.Open(output_file_kb, 'w')
  for entry in data:
    f_kb.write(flatten_json(entry['kb']) + '\n')
    new_arr = []
    if alt_infer:
      new_arr = [entry['intent'], entry['dialogue'].replace('<eod> ', '')]
    else:
      new_arr = [
          entry['intent'], entry['action'], entry['dialogue'],
          entry['boundaries1']
      ]
      # only boundary1 is used but not 2 because it's not necessary.
    f_data.write('|'.join(new_arr) + '\n')
  f_data.close()
  f_kb.close()


# this needs to be fixed..., turns are randomly selected right now.
def write_completion(data, output_file_data_src, output_file_data_tar,
                     output_file_kb):
  """This function write both kb and main data into the files."""
  f_data_src = gfile.Open(output_file_data_src, 'w')
  f_data_tar = gfile.Open(output_file_data_tar, 'w')
  f_kb = gfile.Open(output_file_kb, 'w')
  for entry in data:
    bd1 = entry['boundaries1'].split(' ')
    bd2 = entry['boundaries2'].split(' ')
    start = bd1[0:len(bd1) // 2] + bd2[0:len(bd2) // 2]
    end = bd1[len(bd1) // 2:] + bd2[len(bd2) // 2:]
    # random_turn = random.randint(0, len(start) - 1)
    for random_turn in range(len(start)):
      f_kb.write(flatten_json(entry['kb']) + '\n')
      # print len(start),len(end),len(bd),random_turn
      turn_start = int(start[random_turn])
      turn_end = int(end[random_turn])
      dialogue_split = entry['dialogue'].split(' ')
      # print turn_start,turn_end
      dialogue_src = dialogue_split[0:turn_start + 1]
      dialogue_tar = dialogue_split[turn_start + 1:turn_end + 1]
      src_arr = [entry['intent'], ' '.join(dialogue_src)]
      f_data_src.write('|'.join(src_arr) + '\n')
      tar_arr = [entry['action'], ' '.join(dialogue_tar)]
      f_data_tar.write('|'.join(tar_arr) + '\n')

  f_data_src.close()
  f_data_tar.close()
  f_kb.close()


def write_self_play(data, output_file_data, output_file_kb):
  f_data = gfile.Open(output_file_data, 'w')
  f_kb = gfile.Open(output_file_kb, 'w')
  for entry in data:
    f_kb.write(flatten_json(entry['kb']) + '\n')
    new_arr = [entry['intent'],
               entry['expected_action']]  # intent and action are both needed
    f_data.write('|'.join(new_arr) + '\n')
  f_data.close()
  f_kb.close()


def word_tokenize(tokens):
  return [
      token.replace("''", '"').replace('``', '"')
      for token in nltk.word_tokenize(tokens)
  ]



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""library to standardize data."""

from tensorflow.compat.v1.io import gfile
from tqdm import tqdm
import string
import json

printable = set(string.printable)


def add_dot(utt):
  if utt.strip()[-1] != '.' and utt.strip()[-1] != '?':
    return utt.strip() + '.'
  else:
    return utt.strip()


def standardize_message(utterances, time_stamp=None):
  """this function combines adjacent utternaces that belong to the same talker

     into one. Sometimes time_stamp could be None.
     For example
        <t1> how are you. <t2> I am good. <t2> And you? <eod> <t1>
     will be combined into
        <t1> how are you. <t2> I am good. And you? <eod> <t1>
  """
  new_utterance = []
  new_time_stamp = []
  for i, utt in enumerate(utterances):
    if len(utt.strip()) == 0:
      continue
    utts = utt.split(':')
    talker = utts[0]
    sentence = ':'.join(utts[1:]).strip()
    if len(sentence) == 0:
      continue
    if len(new_utterance) == 0 or talker != new_utterance[-1].split(':')[0]:
      new_utterance.append(add_dot(utt))
      if time_stamp:
        new_time_stamp.append(time_stamp[i])
    else:
      new_utterance[-1] += ' ' + add_dot(sentence)
      if time_stamp:
        new_time_stamp[-1] = time_stamp[i]
  return new_utterance, new_time_stamp


def delete_non_ascii(s):
  return ''.join([x for x in s if x in printable])


def load_and_drop(data_file, kb_file, drop_incorrect=True, verbose=False):
  """ this function filter incorrect samples without standardization."""
  fin_data = gfile.GFile(data_file)
  fin_kb = gfile.GFile(kb_file)
  total_in_file = 0
  loaded_data = []
  loaded_kb = []
  for line1 in tqdm(fin_data, desc='loading data'):
    if len(line1.strip()) < 10:
      continue
    line2 = fin_kb.readline()
    if len(line2.strip()) < 10:
      continue
    line1 = delete_non_ascii(line1)
    line2 = delete_non_ascii(line2)

    data_obj = json.loads(line1)
    kb_obj = json.loads(line2)
    if (not drop_incorrect) or (
        'correct_sample' not in data_obj) or data_obj['correct_sample']:
      loaded_data.append(data_obj)
      loaded_kb.append(kb_obj)
    total_in_file += 1

  if verbose:
    print(('loaded: ', len(loaded_data), '/', total_in_file, '=',
           len(loaded_data) * 1.0 / total_in_file))
  return loaded_data, loaded_kb


def load_and_drop_stream(data_file,
                         kb_file,
                         drop_incorrect=True,
                         verbose=False):
  """ this function filter incorrect samples without standardization."""
  if verbose:
    print('loading stream')
  fin_data = gfile.GFile(data_file)
  if gfile.exists(kb_file):
    fin_kb = gfile.GFile(kb_file)
  else:
    fin_kb = None
  if verbose:
    print('gfile loaded: ', fin_data)
  for line1 in fin_data:
    if verbose:
      print(line1)
    if len(line1.strip()) < 10:
      continue
    line1 = delete_non_ascii(line1)
    data_obj = json.loads(line1)

    if fin_kb:
      line2 = fin_kb.readline()
      if len(line2.strip()) < 10:
        continue
      line2 = delete_non_ascii(line2)
      kb_obj = json.loads(line2)
    else:
      kb_obj = None
    if (not drop_incorrect) or (
        'correct_sample' not in data_obj) or data_obj['correct_sample']:
      yield data_obj, kb_obj


def standardize_and_drop(data_file,
                         kb_file,
                         drop_incorrect=True,
                         verbose=False):
  """ this function filter incorrect samples and standardize them

   the same time.
  """
  loaded_data, loaded_kb = load_and_drop(data_file, kb_file, drop_incorrect,
                                         verbose)
  for data_obj in tqdm(loaded_data, desc='standardizing data'):
    org_time = data_obj['timestamps'] if 'timestamps' in data_obj else None
    org_diag = data_obj['dialogue'] if 'dialogue' in data_obj else None
    if org_diag:
      new_diag, new_time = standardize_message(org_diag, org_time)
      data_obj['dialogue'] = new_diag
      if new_time:
        data_obj['timestamps'] = new_time
        assert len(data_obj['dialogue']) == len(data_obj['timestamps'])
  return loaded_data, loaded_kb



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains the structure to model customers."""
# import ast
import random
from . import utils


class Customer(object):
  """This class contains information of a customer."""

  def __init__(self, facts_obj, book_window, airport_list):
    # 1. origin and destination, airport_list guarantees to have unique locations
    self.origin = random.randint(0, len(airport_list) - 1)
    self.dest = random.randint(0, len(airport_list) - 1)
    if self.dest == self.origin:
      self.dest = (self.dest + 1) % len(airport_list)
    self.dest = airport_list[self.dest]
    self.origin = airport_list[self.origin]
    # 2. date
    base_time = facts_obj.base_departure_time_epoch
    a_year_from_now = base_time + 3600 * 24 * 365
    # randomly pick a date between base_time and a_year_from_now
    self.departure_date = random.randint(base_time, a_year_from_now)
    # return adte is book_window away from the departure date
    self.return_date = self.departure_date + 3600 * 24 * book_window
    # 4. passenger information
    num_passengers = 1
    self.passengers = []
    len_first_name = len(facts_obj.first_name_list)
    len_last_name = len(facts_obj.last_name_list)
    # '_' will later be replaced in intent standalization
    for _ in range(num_passengers):
      self.passengers.append(
          facts_obj.first_name_list[random.randint(0, len_first_name - 1)] +
          '_' + facts_obj.last_name_list[random.randint(0, len_last_name - 1)])
    # non-required fields during initial query
    # 3. time
    self.departure_time = utils.choice(
        facts_obj.time_list, 1, p=facts_obj.time_prior)[0]
    self.return_time = utils.choice(
        facts_obj.time_list, 1, p=facts_obj.time_prior)[0]

    # 5. class limit and price limit
    self.class_limit = utils.choice(
        facts_obj.class_list, 1, p=facts_obj.class_list_prior)[0]

    # 6. price limist
    if self.class_limit == 'all':
      self.price_limit = facts_obj.price_limit_list[random.randint(
          0,
          len(facts_obj.price_limit_list) - 1)]
    elif self.class_limit == 'economy':
      self.price_limit = facts_obj.price_limit_list[random.randint(
          0,
          len(facts_obj.price_limit_list) - 2)]
    elif self.class_limit == 'business':
      self.price_limit = facts_obj.price_limit_list[random.randint(
          1,
          len(facts_obj.price_limit_list) - 1)]
    # 7. num of connections
    self.max_connection = utils.choice(
        facts_obj.connection_member, 1, p=facts_obj.connection_prior)[0]

    # 8. airline preference
    self.airline = utils.choice(
        facts_obj.airline_preference, 1,
        p=facts_obj.airline_preference_prior)[0]
    # 10 post process
    self.departure_month, self.departure_day = utils.get_month_and_day(
        facts_obj, self.departure_date)
    self.return_month, self.return_day = utils.get_month_and_day(
        facts_obj, self.return_date)
    # 11 change reservation
    self.goal = utils.choice([0, 1, 2], p=facts_obj.goal_probaility)

  def get_departure_and_return_date(self):
    return self.departure_date, self.return_date


  def get_json(self):
    """This function serializes the object into a json."""
    intention_jobject = {}
    intention_jobject['departure_airport'] = self.origin
    intention_jobject['return_airport'] = self.dest
    intention_jobject['departure_month'] = self.departure_month
    intention_jobject['departure_day'] = self.departure_day
    intention_jobject['return_month'] = self.return_month
    intention_jobject['return_day'] = self.return_day
    intention_jobject['name'] = self.passengers[0]
    intention_jobject['departure_time'] = self.departure_time
    intention_jobject['return_time'] = self.return_time
    intention_jobject['class'] = self.class_limit
    intention_jobject['max_price'] = self.price_limit
    intention_jobject['max_connections'] = self.max_connection
    intention_jobject['airline_preference'] = self.airline
    intention_jobject['goal'] = self.goal

    # add departure and return date
    intention_jobject['departure_date'] = self.departure_date
    intention_jobject['return_date'] = self.return_date
    return intention_jobject

  def get_customer_condition(self):
    """This function returns the condition file."""
    condition = self.get_json()
    if condition['airline_preference'] == 'all':
      del condition['airline_preference']
    if condition['max_connections'] == 2:
      del condition['max_connections']
    if condition['class'] == 'all':
      del condition['class']
    if condition['departure_time'] == 'all':
      del condition['departure_time']
    if condition['return_time'] == 'all':
      del condition['return_time']
    return condition




# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This is the entry of the flight simulator generator."""



import copy
import json
import random
import numpy as np
from tensorflow.compat.v1 import gfile
from airdialogue.context_generator.src import customer
from airdialogue.context_generator.src import facts
from airdialogue.context_generator.src import kb as knowledgebase
from airdialogue.context_generator.src import utils


class ContextGenerator(object):
  """context generator."""

  def __init__(self, num_candidate_airports, book_window, num_db_record,
               firstname_file, lastname_file, airportcode_file):
    self.fact_obj = facts.Facts(firstname_file, lastname_file, airportcode_file)
    # number of airports to be considered when generating database.
    self.num_candidate_airports = num_candidate_airports
    # number of days between departuer and arrival
    self.book_window = book_window
    # number of database records
    self.num_db_record = num_db_record

  def _generate_user_action(self, expected_action):
    user_action = copy.deepcopy(expected_action)
    if len(user_action["flight"]) != 0:
      idx = random.randint(0, len(user_action["flight"]) - 1)
      single_flight = user_action["flight"][idx]
      user_action["flight"] = [single_flight]
    return user_action

  def generate_context(self,
                       num_context,
                       output_data=None,
                       output_kb=None,
                       display_freq=None,
                       output_object=False,
                       verbose=False):
    """generate context. if output_file is not none then we write to file."""
    if output_data and output_kb:
      fp_data = gfile.Open(output_data, "w")
      fp_kb = gfile.Open(output_kb, "w")
    else:
      fp_data = None
      fp_kb = None
    all_context = []
    status_stats = {}
    for n in range(num_context):
      airport_candidate = list(
          np.random.choice(
              self.fact_obj.airport_list,
              self.num_candidate_airports,
              replace=False))
      if display_freq:
        if verbose and n % display_freq == 0:
          print((n, "/", num_context))
      cus = customer.Customer(self.fact_obj, self.book_window,
                              airport_candidate)
      ref_departure_date, ref_return_date = cus.departure_date, cus.return_date
      kb = knowledgebase.Knowledgebase(self.fact_obj, self.num_db_record,
                                       airport_candidate, ref_departure_date,
                                       ref_return_date)
      intent_json = utils.standardize_intent(cus.get_json())
      kb_and_res_json = kb.get_json()
      kb_json = kb_and_res_json["kb"]
      res_json = kb_and_res_json["reservation"]
      expected_action = utils.generate_expected_action(
          self.fact_obj, intent_json, kb_json, res_json)
      status = expected_action["status"]
      if status not in status_stats:
        status_stats[status] = 0
      status_stats[status] += 1

      if fp_data and fp_kb:
        context_data = {
            "intent": intent_json,
            "action": self._generate_user_action(expected_action),
            "expected_action": expected_action
        }
        context_kb = {"kb": kb_json, "reservation": res_json}

        fp_data.write(json.dumps(context_data) + "\n")
        fp_kb.write(json.dumps(context_kb) + "\n")
      else:
        if output_object:  # output pythoh object instead of json
          object_structure = (cus, kb, expected_action)
          all_context.append(object_structure)
        else:  # output json
          context_json = {
              "intent": intent_json,
              "kb": kb_json,
              "reservation": res_json,
              "action": self._generate_user_action(expected_action),
              "expected_action": expected_action
          }
          all_context.append(context_json)

    if fp_data:
      fp_data.close()
    if fp_kb:
      fp_kb.close()
    # status stats
    for key in status_stats:
      status_stats[key] /= 1.0 * num_context

    return all_context, status_stats



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from airdialogue.evaluator.metrics.flight_distance import generate_scaled_flight
from airdialogue.evaluator.metrics import f1


def compute_reward_batch(utterance,
                         predicted_action,
                         actual_action_concat,
                         flight_db,
                         hparams,
                         alpha=0.5,
                         beta=0.2,
                         gamma=0.3):
  """Calcualte the reward for a batch."""
  rewards = []
  acc1 = []
  acc2 = []
  acc3 = []
  discrete_score = []
  ds1_arr = []
  ds2_arr = []
  ds3_arr = []
  train_rw_arr = []
  for pa, aa_con, fl in zip(predicted_action, actual_action_concat, flight_db):
    aa = aa_con.split(' ')
    rw, ac1, ac2, ac3 = compute_reward(pa, aa, fl)
    rewards.append(rw)
    acc1.append(ac1)
    acc2.append(ac2)
    acc3.append(ac3)

    ds, ds1, ds2, ds3 = compute_01_score(pa, aa)
    discrete_score.append(ds)
    ds1_arr.append(ds1)
    ds2_arr.append(ds2)
    ds3_arr.append(ds3)
    train_rw_arr.append(
        get_training_reward(hparams, ac1, ac2, ac3, ds1, ds2, ds3))
  return train_rw_arr, rewards, acc1, acc2, acc3, discrete_score, ds1_arr, ds2_arr, ds3_arr


def parse_action(action):
  """parse the action and consider multiple name scenario.
  name will also appear first.
  """
  name = ' '.join(action[0:-2])
  flight = action[-2]
  state = action[-1]
  return name, flight, state


def compute_reward(predicted_action,
                   actual_action,
                   flight_db,
                   alpha=0.5,
                   beta=0.2,
                   gamma=0.3,
                   debug=False):
  """here we compute the scaled reward."""
  predicted_name, predicted_flight, predicted_state = parse_action(
      predicted_action)
  actual_name, actual_flight, actual_state = parse_action(actual_action)

  # this will do normalization including lower case and prouncation/space
  # removal
  score1 = f1.f1_score(predicted_name, actual_name)
  score2 = 1 - generate_scaled_flight(predicted_flight, actual_flight,
                                      flight_db)
  score3 = float(predicted_state == actual_state)

  reward_compliment = score1 * 0.2 + score2 * 0.5 + score3 * 0.3

  acc1 = score1
  acc2 = score2
  acc3 = score3
  return reward_compliment, acc1, acc2, acc3





# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python implementation of BLEU and smooth-BLEU.

This module provides a Python implementation of BLEU and smooth-BLEU.
Smooth BLEU is computed following the method outlined in the paper:
Chin-Yew Lin, Franz Josef Och. ORANGE: a method for evaluating automatic
evaluation metrics for machine translation. COLING 2004.
"""

import collections
import math
import tqdm

from multiprocessing import Pool
from functools import partial


def _get_ngrams(segments, order):
  """Extracts all n-grams upto a given maximum order from an input segment.

  Args:
    segments: text segment from which n-grams will be extracted.
    order: n of the n-grams returned.

  Returns:
    The Counter containing all n-grams in segment
    with a count of how many times each n-gram occurred.
  """
  ngram_counts = collections.Counter()
  for seg in segments:
    for i in range(0, len(seg) - order + 1):
      ngram = tuple(seg[i:i+order])
      ngram_counts[ngram] += 1
  return ngram_counts


def compute_kl(reference_corpus,
               translation_corpus,
               max_order=4,
               freq_thre=100,
               workers=20,
               verbose=False):
  """Computes KLdivergence of translated segments against one or more references.

  Args:
    reference_corpus: list of lists of references for each translation. Each
        reference should be tokenized into a list of tokens.
    translation_corpus: list of translations to score. Each translation
        should be tokenized into a list of tokens.
    max_order: Maximum n-gram order to use when computing BLEU score.
    smooth: Whether or not to apply Lin et al. 2004 smoothing.

  Returns:
    3-Tuple with the BLEU score, n-gram precisions, geometric mean of n-gram
    precisions and brevity penalty.
  """
  results = {}
  itertool = tqdm.tqdm if verbose else lambda x, *args, **kw: x

  for order in range(1, max_order + 1):
    with Pool(processes=workers) as pool:
      chunksize = len(reference_corpus)//(3*workers)
      chunkedexamples = [reference_corpus[i:i + chunksize]
                         for i in range(0, len(reference_corpus), chunksize)]
      ngram_counts = list(itertool(pool.imap(partial(_get_ngrams, order=order), chunkedexamples),\
                                    total=len(chunkedexamples), \
                                    desc="Calculating {}-gram for ref data".format(order)))
      merged_ref_ngram_counts = collections.Counter()
      for c in ngram_counts:
        merged_ref_ngram_counts += c
    with Pool(processes=workers) as pool:
      chunksize = len(translation_corpus)//(10*workers)
      chunkedexamples = [translation_corpus[i:i + chunksize]
                         for i in range(0, len(translation_corpus), chunksize)]
      ngram_counts = list(itertool(pool.imap(partial(_get_ngrams, order=order), chunkedexamples),\
                                    total=len(chunkedexamples), \
                                    desc="Calculating {}-gram for trans data".format(order)))
      merged_trans_ngram_counts = collections.Counter()
      for c in ngram_counts:
        merged_trans_ngram_counts += c

    merged_ref_ngram_counts = \
      collections.Counter({k: c for k, c in merged_ref_ngram_counts.items() if c >= freq_thre})
    ref_total_nums = sum([c for k, c in merged_ref_ngram_counts.items()])
    # We do not remove items from merged_trans_ngram_counts to prevent infite issue
    trans_total_nums = sum([c for k, c in merged_trans_ngram_counts.items() if c >= freq_thre])

    kl = 0
    # eps: smoothing parameter
    eps = 1e-10
    for k, c in merged_ref_ngram_counts.items():
      _c = merged_trans_ngram_counts.get(k, 0)

      # apply smoothing
      c += eps
      _c += eps

      kl += - c/ref_total_nums * math.log((_c/c)*(ref_total_nums/trans_total_nums))

    results["{}-gram".format(order)] = kl

  return results



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility for evaluating various tasks."""
import codecs
import tensorflow.compat.v1 as tf

from airdialogue.evaluator.metrics import bleu
from airdialogue.evaluator.metrics import rouge
from airdialogue.evaluator.metrics import kl

ROLE_TOKENS = ["<t1>", "<t2>"]


def evaluate(ref_file, trans_file, metric):
  """Pick a metric and evaluate depending on task."""
  if ":" in metric:
    metric, mode = metric.split(":")
  else:
    mode = "brief"
  assert mode in ["brief", "all"]
  # BLEU scores for translation task
  if metric.lower() == "bleu":
    evaluation_score = _bleu(
        ref_file, trans_file, mode=mode)
  # ROUGE scores for summarization tasks
  elif metric.lower() == "rouge":
    evaluation_score = _rouge(
        ref_file, trans_file, mode=mode)
  # kl scores for evaluating the ngram kl distribution of the whole corpus
  elif metric.lower() == "kl":
    evaluation_score = _kl(
        ref_file, trans_file, mode=mode)
  elif metric.lower() == "accuracy":
    evaluation_score = _accuracy(ref_file, trans_file)
  else:
    raise ValueError("Unknown metric %s" % metric)

  return evaluation_score

def _kl(ref_file, trans_file, mode="brief"):
  """Compute KL divergence and handling BPE."""
  max_order = 4

  ref_files = [ref_file]
  reference_text = []
  role_tokens = []
  for reference_filename in ref_files:
    with codecs.getreader("utf-8")(tf.gfile.GFile(reference_filename,
                                                  "rb")) as fh:
      for line in fh:
        reference, role = process_dialogue_infer(
            line.rstrip(), get_role_token=True)
        reference_text.append(reference.split(" "))
        role_tokens.append(role)

  translations = []
  with codecs.getreader("utf-8")(tf.gfile.GFile(trans_file, "rb")) as fh:
    for line in fh:
      translations.append(line.rstrip().split(" "))

  results = {}
  kl_scores = kl.compute_kl(reference_text, translations, max_order)
  for key in kl_scores:
    results["all-" + key] = kl_scores[key]
  if mode == "brief":
    return sum(results.values()) / len(results)

  for role in ROLE_TOKENS:
    _sub_ref_texts = []
    _sub_trans = []
    for _r, _t, _role in zip(reference_text, translations, role_tokens):
      if _role == role:
        _sub_ref_texts.append(_r)
        _sub_trans.append(_t)
    kl_scores = kl.compute_kl(_sub_ref_texts, _sub_trans, max_order)
    for key in kl_scores:
      results[role + "-" + key] = kl_scores[key]
  return results

def _bleu(ref_file, trans_file, mode="brief"):
  """Compute BLEU scores and handling BPE."""
  max_order = 4
  smooth = False

  ref_files = [ref_file]
  reference_text = []
  for reference_filename in ref_files:
    with codecs.getreader("utf-8")(tf.gfile.GFile(reference_filename,
                                                  "rb")) as fh:
      reference_text.append(fh.readlines())

  per_segment_references = []
  role_tokens = []
  for references in zip(*reference_text):
    reference_list = []
    for reference in references:
      reference, role = process_dialogue_infer(
          reference.rstrip(), get_role_token=True)
      reference_list.append(reference.split(" "))
    per_segment_references.append(reference_list)
    role_tokens.append(role)

  translations = []
  with codecs.getreader("utf-8")(tf.gfile.GFile(trans_file, "rb")) as fh:
    for line in fh:
      translations.append(line.rstrip().split(" "))

  results = {}
  bleu_score, _, _, _, _, _ = bleu.compute_bleu(per_segment_references,
                                                translations, max_order, smooth)
  results["all"] = 100 * bleu_score
  if mode == "brief":
    return results["all"]

  for role in ROLE_TOKENS:
    _sub_ref_texts = []
    _sub_trans = []
    for _r, _t, _role in zip(per_segment_references, translations, role_tokens):
      if _role == role:
        _sub_ref_texts.append(_r)
        _sub_trans.append(_t)
    bleu_score, _, _, _, _, _ = bleu.compute_bleu(_sub_ref_texts, _sub_trans,
                                                  max_order, smooth)
    results[role] = 100 * bleu_score

  return results


def _rouge(ref_file, summarization_file, mode="brief"):
  """Compute ROUGE scores and handling BPE."""

  results = {}

  references = []
  role_tokens = []
  with codecs.getreader("utf-8")(tf.gfile.GFile(ref_file, "rb")) as fh:
    for line in fh:
      ref, role = process_dialogue_infer(line.rstrip(), get_role_token=True)
      references.append(ref)
      role_tokens.append(role)

  hypotheses = []
  with codecs.getreader("utf-8")(tf.gfile.GFile(summarization_file,
                                                "rb")) as fh:
    for line in fh:
      hypotheses.append(line)

  rouge_score_map = rouge.rouge(hypotheses, references)
  results["all"] = 100 * rouge_score_map["rouge_l/f_score"]
  if mode == "brief":
    return results["all"]

  for role in ROLE_TOKENS:
    _sub_ref_texts = []
    _sub_hypos = []
    for _r, _t, _role in zip(references, hypotheses, role_tokens):
      if _role == role:
        _sub_ref_texts.append(_r)
        _sub_hypos.append(_t)
    rouge_score_map = rouge.rouge(_sub_hypos, _sub_ref_texts)
    results[role] = 100 * rouge_score_map["rouge_l/f_score"]

  return results


def process_dialogue_infer(file_line, get_role_token=False):
  # split the end token (<t1>,<t2>)
  _line = file_line.replace(" <eod>", "")
  _line = _line.rstrip().split("|")[1].rsplit(" ", 1)
  if not get_role_token:
    return _line[0]
  else:
    return _line[0], _line[1]


def _accuracy(label_file, pred_file):
  """Compute accuracy, each line contains a label."""

  with codecs.getreader("utf-8")(tf.gfile.GFile(label_file, "rb")) as label_fh:
    with codecs.getreader("utf-8")(tf.gfile.GFile(pred_file, "rb")) as pred_fh:
      count = 0.0
      match = 0.0
      for label, pred in zip(label_fh, pred_fh):
        label = process_dialogue_infer(label.strip()).strip()
        pred = pred.strip()
        if label == pred:
          match += 1
        count += 1
  return 100 * match / count



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module builds a falsk server for visualization."""

import argparse
import json
import linecache
import os
from os.path import expanduser
from flask import Flask
from flask import request
from airdialogue.visualizer.utils import generate_html
import sys

def strip_prefix(name):
  if name.endswith("data"):
    return name[0:-5]
  else:
    return name[0:-3]


def get_partitions(path):
  """This function counts the number of occurries of the same prefix in the
     json dir. If it happens more than twice we will use that as a valid
     partition and add it to the partition list."""
  all_files = os.listdir(path)
  prefix_freq = {}
  for f in all_files:
    if f.endswith(".json"):
      prefix = f.split(".")[0]
      prefix = strip_prefix(prefix)
      if prefix not in prefix_freq:
        prefix_freq[prefix] = 0
      prefix_freq[prefix] += 1
  valid_partitions = []
  for prefix in prefix_freq:
    if prefix_freq[prefix] >= 2:
      valid_partitions.append(prefix)
  return valid_partitions


def wrapper(FLAGS):
  def home():
    # get all the partitions in the directory
    expanded_data_path = expanduser(FLAGS.data_path)
    partitions = get_partitions(expanded_data_path)
    index = request.form.get("index")
    if index:
      index = int(index)
    partition = request.form.get("partition")

    if not partitions:
      print("no data is found in the directory")
      return """No partitions found under {0}. Supported partitions has to end
                with .json extension.""".format(FLAGS.data_path)
    if not partition:
      # choose a default partition
      if "train" in partitions:
        partition = "train"
      else:
        partition = partitions[0].strip()
      index = 1

    try:
      line_data = linecache.getline(
          os.path.join(expanded_data_path, "{0}_data.json".format(partition)),
          index)
      line_kb = linecache.getline(
          os.path.join(expanded_data_path, "{0}_kb.json".format(partition)),
          index)
    except:
      return "Invalid index."

    if (not line_data) and (not line_kb):
      print("invalid partition number.")
    data_object = json.loads(line_data)
    kb_object = json.loads(line_kb)
    html_source = generate_html(data_object, kb_object, index, partitions,
                                partition)
    return html_source
  return home


def add_arguments(parser):
  """Build ArgumentParser."""
  parser.add_argument("--host", type=str, default="0.0.0.0",
                      help="host name for the visualizer.")
  parser.add_argument("--port", type=int, default=5555,
                      help="port number of the visualizer server.")
  parser.add_argument("--data_path", type=str, default=None,
                      help="path that stores data and kb files.")


def main(FLAGS):
  app = Flask(__name__)
  app.route("/", methods=["POST", "GET"])(wrapper(FLAGS))
  app.run(host=FLAGS.host, port=FLAGS.port, debug=True)


if __name__ == "__main__":
  this_parser = argparse.ArgumentParser()
  add_arguments(this_parser)
  FLAGS, unparsed = this_parser.parse_known_args()
  main(FLAGS)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This is the main module that generates contexts."""

import argparse
from airdialogue.context_generator import context_generator_lib
import tensorflow.compat.v1 as tf
import sys

FLAGS= None

def add_arguments(parser):
  """Build ArgumentParser."""
  parser.register("type", "bool", lambda v: v.lower() == "true")
  parser.add_argument(
      "--num_candidate_airports",
      type=int,
      default=3,
      help="number of candidate airports")
  parser.add_argument(
      "--book_window", type=int, default=2, help="window between bookings")
  parser.add_argument(
      "--num_db_record",
      type=int,
      default=30,
      help="number of database records per booking.")
  parser.add_argument(
      "--firstname_file",
      type=str,
      default="./data/resources/meta_context/first_names.txt",
      help="text file that contains a list of first names.")
  parser.add_argument(
      "--lastname_file",
      type=str,
      default="./data/resources/meta_context/last_names.txt",
      help="text file that contains a list of last names.")
  parser.add_argument(
      "--airportcode_file",
      type=str,
      default="./data/resources/meta_context/airport.txt",
      help="text file that contains a list of airport codes.")
  parser.add_argument(
      "--num_samples",
      type=int,
      default=320000,
      help="number of samples to generate.")
  parser.add_argument(
      "--output_data",
      type=str,
      default=None,
      help="path of the output data file.")
  parser.add_argument(
      "--output_kb", type=str, default=None, help="path of the output kb file.")
  parser.add_argument(
      "--display_freq",
      type=int,
      default=50000,
      help="display frequency for information.")
  parser.add_argument(
      "--verbose",
      type="bool",
      nargs="?",
      const=True,
      default=False,
      help="if enabled, debug info will be printed out.")


def main(FLAGS):
  cg = context_generator_lib.ContextGenerator(
      num_candidate_airports=FLAGS.num_candidate_airports,
      book_window=FLAGS.book_window,
      num_db_record=FLAGS.num_db_record,
      firstname_file=FLAGS.firstname_file,
      lastname_file=FLAGS.lastname_file,
      airportcode_file=FLAGS.airportcode_file)

  _, stats = cg.generate_context(FLAGS.num_samples,
                                 output_data=FLAGS.output_data,
                                 output_kb=FLAGS.output_kb,
                                 display_freq=FLAGS.display_freq,
                                 verbose=FLAGS.verbose)
  if FLAGS.verbose: print(stats)


def run_main(unused):
  main(FLAGS)

if __name__ == "__main__":
  this_parser = argparse.ArgumentParser()
  add_arguments(this_parser)
  FLAGS, unparsed = this_parser.parse_known_args()
  tf.app.run(main=run_main, argv=[sys.argv[0]] + unparsed)





# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python implementation of BLEU and smooth-BLEU.

This module provides a Python implementation of BLEU and smooth-BLEU.
Smooth BLEU is computed following the method outlined in the paper:
Chin-Yew Lin, Franz Josef Och. ORANGE: a method for evaluating automatic
evaluation metrics for machine translation. COLING 2004.
"""

import collections
import math


def _get_ngrams(segment, max_order):
  """Extracts all n-grams upto a given maximum order from an input segment.

  Args:
    segment: text segment from which n-grams will be extracted.
    max_order: maximum length in tokens of the n-grams returned by this
        methods.

  Returns:
    The Counter containing all n-grams upto max_order in segment
    with a count of how many times each n-gram occurred.
  """
  ngram_counts = collections.Counter()
  for order in range(1, max_order + 1):
    for i in range(0, len(segment) - order + 1):
      ngram = tuple(segment[i:i+order])
      ngram_counts[ngram] += 1
  return ngram_counts


def compute_bleu(reference_corpus, translation_corpus, max_order=4,
                 smooth=False):
  """Computes BLEU score of translated segments against one or more references.

  Args:
    reference_corpus: list of lists of references for each translation. Each
        reference should be tokenized into a list of tokens.
    translation_corpus: list of translations to score. Each translation
        should be tokenized into a list of tokens.
    max_order: Maximum n-gram order to use when computing BLEU score.
    smooth: Whether or not to apply Lin et al. 2004 smoothing.

  Returns:
    3-Tuple with the BLEU score, n-gram precisions, geometric mean of n-gram
    precisions and brevity penalty.
  """
  matches_by_order = [0] * max_order
  possible_matches_by_order = [0] * max_order
  reference_length = 0
  translation_length = 0
  for (references, translation) in zip(reference_corpus,
                                       translation_corpus):
    reference_length += min(len(r) for r in references)
    translation_length += len(translation)

    merged_ref_ngram_counts = collections.Counter()
    for reference in references:
      merged_ref_ngram_counts |= _get_ngrams(reference, max_order)
    translation_ngram_counts = _get_ngrams(translation, max_order)
    overlap = translation_ngram_counts & merged_ref_ngram_counts
    for ngram in overlap:
      matches_by_order[len(ngram)-1] += overlap[ngram]
    for order in range(1, max_order+1):
      possible_matches = len(translation) - order + 1
      if possible_matches > 0:
        possible_matches_by_order[order-1] += possible_matches

  precisions = [0] * max_order
  for i in range(0, max_order):
    if smooth:
      precisions[i] = ((matches_by_order[i] + 1.) /
                       (possible_matches_by_order[i] + 1.))
    else:
      if possible_matches_by_order[i] > 0:
        precisions[i] = (float(matches_by_order[i]) /
                         possible_matches_by_order[i])
      else:
        precisions[i] = 0.0

  if min(precisions) > 0:
    p_log_sum = sum((1. / max_order) * math.log(p) for p in precisions)
    geo_mean = math.exp(p_log_sum)
  else:
    geo_mean = 0

  ratio = float(translation_length) / reference_length

  if ratio <= 0.0:
    bp = 0.
  elif ratio > 1.0:
    bp = 1.
  else:
    bp = math.exp(1 - 1. / ratio)

  bleu = geo_mean * bp

  return (bleu, precisions, bp, ratio, translation_length, reference_length)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This file only tokenize one file at a time."""

import argparse
import os
import nltk
from tensorflow.compat.v1 import gfile
import tensorflow.compat.v1 as tf
from tqdm import tqdm

from airdialogue.prepro.tokenize_lib import list_of_action_tokens_except_name
from airdialogue.prepro.tokenize_lib import process_kb
from airdialogue.prepro.tokenize_lib import process_main_data
from airdialogue.prepro.tokenize_lib import word_tokenize
from airdialogue.prepro.tokenize_lib import write_cat
from airdialogue.prepro.tokenize_lib import write_completion
from airdialogue.prepro.tokenize_lib import write_data
from airdialogue.prepro.tokenize_lib import write_self_play
from airdialogue.prepro.tokenize_lib import write_vocabulary
# Standardization libs
from airdialogue.prepro.standardize_data_lib import standardize_and_drop
from airdialogue.prepro.standardize_data_lib import load_and_drop, load_and_drop_stream
import sys
FLAGS = None


def add_arguments(parser):
  """Build ArgumentParser."""
  parser.register('type', 'bool', lambda v: v.lower() == 'true')
  parser.add_argument(
      '--word_cutoff', type=int, default=0, help='number of candidate airports')
  parser.add_argument(
      '--data_file', type=str, default=None, help='path for data_file')
  parser.add_argument(
      '--kb_file', type=str, default=None, help='path for kb_file')
  parser.add_argument(
      '--output_prefix',
      type=str,
      default='train',
      help='prefix of output file')
  parser.add_argument('--output_dir', type=str, default=None, help='output dir')
  parser.add_argument(
      '--input_type', type=str, default='dialogue', help='dialogue/context')
  parser.add_argument(
      '--job_type',
      type=str,
      default='1|1|1|0|0',
      help='train|eval|infer|sp-train|sp-eval')
  parser.add_argument(
      '--nltk_data', type=str, default=None, help='path to NLTK data')
  parser.add_argument(
      '--verbose',
      type='bool',
      nargs='?',
      const=True,
      default=False,
      help='if enabled, debug info will be printed out.')
  parser.add_argument(
      '--skip_standardize',
      type='bool',
      nargs='?',
      const=True,
      default=False,
      help="""if enabled, dialogues will skip the
                              standardization process, which includes
                              removeing duplicated end of sentence tokens and
                              filtering out non-ascii characters.""")
  parser.add_argument(
      '--keep_incorrect',
      type='bool',
      nargs='?',
      const=True,
      default=False,
      help="""if enabled, incorrect dialogues will not be
                              filtered out. In stead, they will be generated
                              along with the correct ones. """)
  parser.add_argument(
      '--gen_cat',
      type='bool',
      nargs='?',
      const=True,
      default=False,
      help='if enabled, category files will be generated.')
  parser.add_argument(
      '--gen_voc',
      type='bool',
      nargs='?',
      const=True,
      default=False,
      help='if enabled, vocabulary file will be generated.')
  parser.add_argument(
      '--gen_voc_map',
      type='bool',
      nargs='?',
      const=True,
      default=False,
      help='if enabled, vocabulary map with counting will be generated.')
  parser.add_argument(
      '--gen_special_token',
      type='bool',
      nargs='?',
      const=True,
      default=False,
      help='if enabled, special token file will be generated.')
  parser.add_argument(
      '--keep_non_ascii',
      type='bool',
      nargs='?',
      const=True,
      default=False,
      help='if enabled, non-ascii tokens will not be droped in vocabulary')
  parser.add_argument(
      '--infer_src_data_file',
      type=str,
      default=None,
      help='path for infer_src_data_file')
  parser.add_argument(
      '--infer_kb_file', type=str, default=None, help='path for infer_kb_file')
  parser.add_argument(
      '--self_play_start_turn',
      type=str,
      default=None,
      help='[agent|customer] whether agent or customer starts empty conversations'
  )


def generate_entry(intents, actions, expected_actions, dialogues, kbs,
                   boundaries1, boundaries2, i):
  """generate entry based on the available of data."""
  # required fields
  entry = {'intent': intents[i], 'action': actions[i], 'kb': kbs[i]}
  # optional fields depending on the input type
  if dialogues:
    entry['dialogue'] = dialogues[i]
  if boundaries1:
    entry['boundaries1'] = boundaries1[i]
  if boundaries2:
    entry['boundaries2'] = boundaries2[i]
  if expected_actions:
    entry['expected_action'] = expected_actions[i]
  return entry


def reorganize_data(intents, actions, expected_actions, dialogues, kbs,
                    boundaries1, boundaries2):
  """reorganize data into compact form.

  This is mostly because we want to reuse
  code for the random split version.
  """
  data = []
  for i in range(len(intents)):
    entry = generate_entry(intents, actions, expected_actions, dialogues, kbs,
                           boundaries1, boundaries2, i)
    data.append(entry)
  return data


def process_job_type(job_type_str, input_type):
  """process job type string."""
  all_jobs = []
  job_type_arr = job_type_str.split('|')
  assert len(job_type_arr) == 5
  for i in range(5):
    if int(job_type_arr[i]) == 1:
      all_jobs.append('train|eval|infer|sp-train|sp-eval'.split('|')[i])
  assert input_type in ['dialogue', 'context']
  # if input type is context, we can't do train/eval/infer
  if input_type == 'context':
    assert 'train' not in all_jobs
    assert 'eval' not in all_jobs
    assert 'infer' not in all_jobs
  return all_jobs


def load_data_from_jsons(FLAGS, input_data_file, input_kb_file, output_vab,
                         output_all_vab, gen_cat, cat_files):
  vocal_map = {}
  sent_tokenize = nltk.sent_tokenize

  raw_data, raw_kb = load_and_drop(
      input_data_file,
      input_kb_file,
      drop_incorrect=not FLAGS.keep_incorrect,
      verbose=FLAGS.verbose)
  # has to be there no matter what
  if FLAGS.verbose:
    print('processing kb')
  processed_kb, vocal_map = process_kb(raw_kb, vocal_map)
  # if dialogue, everything will be there.
  # if context, only intents, actions, vocal_map will be there
  if FLAGS.verbose:
    print('processing data')
  result = process_main_data(
      raw_data,
      sent_tokenize,
      word_tokenize,
      vocal_map,
      input_type=FLAGS.input_type)
  intents, actions, expected_actions, dialogues, vocal_map, boundaries1, boundaries2, cats = result
  frequency_cutoff = FLAGS.word_cutoff
  # 3 is the number of special tokens
  if FLAGS.verbose:
    print('vocabulary before cutoff', len(vocal_map) + 3)
  vocal_map = write_vocabulary(output_vab, output_all_vab, vocal_map,
                               frequency_cutoff, FLAGS.keep_non_ascii)
  if gen_cat:
    if FLAGS.verbose:
      print('writing category')
    write_cat(cat_files, cats)

  if FLAGS.verbose:
    print(
        'frequency_cutoff= {0}, vocabulary after cutoff'.format(
            frequency_cutoff), len(vocal_map))
  data = reorganize_data(intents, actions, expected_actions, dialogues,
                         processed_kb, boundaries1, boundaries2)
  return data


def load_data_from_jsons_stream(FLAGS,
                                input_data_file,
                                input_kb_file,
                                output_vab,
                                output_all_vab,
                                gen_cat,
                                cat_files,
                                self_play_start_turn=None):
  vocal_map = {}
  sent_tokenize = nltk.sent_tokenize

  for raw_data, raw_kb in tqdm(
      load_and_drop_stream(
          input_data_file,
          input_kb_file,
          drop_incorrect=not FLAGS.keep_incorrect,
          verbose=FLAGS.verbose),
      desc='processing stream'):
    # has to be there no matter what
    if raw_kb is not None:
      processed_kb, vocal_map = process_kb([raw_kb], vocal_map, stream=True)
    else:
      processed_kb = [['no_res']]
    # if dialogue, everything will be there.
    # if context, only intents, actions, vocal_map will be there
    result = process_main_data([raw_data],
                               sent_tokenize,
                               word_tokenize,
                               vocal_map,
                               stream=True,
                               input_type=FLAGS.input_type,
                               self_play_start_turn=self_play_start_turn)
    intents, actions, expected_actions, dialogues, vocal_map, boundaries1, boundaries2, cats = result
    frequency_cutoff = FLAGS.word_cutoff
    # 3 is the number of special tokens
    # if FLAGS.verbose: print 'vocabulary before cutoff', len(vocal_map) + 3
    # vocal_map = write_vocabulary(output_vab, output_all_vab, vocal_map,
    #                             frequency_cutoff, FLAGS.keep_non_ascii)
    # print("CC")
    # print(vocal_map)
    if gen_cat:
      if FLAGS.verbose:
        print('writing category')
      write_cat(cat_files, cats)

    if FLAGS.verbose:
      print(
          'frequency_cutoff= {0}, vocabulary after cutoff'.format(
              frequency_cutoff), len(vocal_map))
    data = reorganize_data(intents, actions, expected_actions, dialogues,
                           processed_kb, boundaries1, boundaries2)[0]
    yield data


def main(FLAGS):
  all_jobs = process_job_type(FLAGS.job_type, FLAGS.input_type)
  output_dir = FLAGS.output_dir
  if FLAGS.verbose:
    print('all_jobs', all_jobs)
    print('input_type', FLAGS.input_type)
    print('output_dir', output_dir)
    print('data_file', FLAGS.data_file)
    print('kb_file', FLAGS.kb_file)
    print('output_prefix', FLAGS.output_prefix)
    print('skip_standardize', FLAGS.skip_standardize)
    print('keep_incorrect', FLAGS.keep_incorrect)
    print('word_cutoff', FLAGS.word_cutoff)
    print('gen_voc', FLAGS.gen_voc)
    print('infer_src_data_file', FLAGS.infer_src_data_file)
    print('infer_kb_file', FLAGS.infer_kb_file)

  if not tf.io.gfile.isdir(output_dir):
    gfile.MkDir(output_dir)

  input_data_file = FLAGS.data_file
  input_kb_file = FLAGS.kb_file
  if len(FLAGS.output_prefix.strip()) == 0:
    FLAGS.output_prefix = ''
  else:
    FLAGS.output_prefix = FLAGS.output_prefix
  # output_vab = output_dir + '/{0}.vocab'.format(FLAGS.output_prefix)
  output_vab = output_dir + '/vocab.txt'
  output_all_vab = output_dir + '/{0}.full.vocab'.format(FLAGS.output_prefix)
  all_token_file = output_dir + '/{0}.special.vocab'.format(FLAGS.output_prefix)
  first_name_cats_file = output_dir + '/{0}.firstname.cat'.format(
      FLAGS.output_prefix)
  last_name_cats_file = output_dir + '/{0}.lastname.cat'.format(
      FLAGS.output_prefix)
  flight_cats_file = output_dir + '/{0}.flight.cat'.format(FLAGS.output_prefix)
  status_cats_file = output_dir + '/{0}.status.cat'.format(FLAGS.output_prefix)
  cat_files = [
      first_name_cats_file, last_name_cats_file, flight_cats_file,
      status_cats_file
  ]

  output_data_pattern = output_dir + '/{0}data'
  output_kb_pattern = output_dir + '/{0}kb'

  nltk_path = FLAGS.nltk_data
  nltk.data.path.append(nltk_path)
  sent_tokenize = nltk.sent_tokenize

  infer_flag_exists = FLAGS.infer_src_data_file or FLAGS.infer_kb_file

  if any(j != 'infer' for j in all_jobs) or not infer_flag_exists:
    # We need to process the default json
    data = load_data_from_jsons(FLAGS, input_data_file, input_kb_file,
                                output_vab, output_all_vab, FLAGS.gen_cat,
                                cat_files)

  if 'infer' in all_jobs and infer_flag_exists:
    # We need to process alternate infer json
    alt_infer_data = load_data_from_jsons_stream(FLAGS,
                                                 FLAGS.infer_src_data_file,
                                                 FLAGS.infer_kb_file, None,
                                                 None, False, [],
                                                 FLAGS.self_play_start_turn)
  if 'train' in all_jobs:
    if FLAGS.verbose:
      print('writing train data')
    write_data(data, output_data_pattern.format(FLAGS.output_prefix + '.'),
               output_kb_pattern.format(FLAGS.output_prefix + '.'))
  if 'eval' in all_jobs:
    if FLAGS.verbose:
      print('writing eval data')
    write_data(data, output_data_pattern.format(FLAGS.output_prefix + '.eval.'),
               output_kb_pattern.format(FLAGS.output_prefix + '.eval.'))
  if 'infer' in all_jobs:
    if FLAGS.verbose:
      print('writing infer data')
    if infer_flag_exists:
      write_data(
          alt_infer_data,
          output_data_pattern.format(FLAGS.output_prefix + '.infer.src.'),
          output_kb_pattern.format(FLAGS.output_prefix + '.infer.'),
          alt_infer=True)
    else:
      write_completion(
          data, output_data_pattern.format(FLAGS.output_prefix + '.infer.src.'),
          output_data_pattern.format(FLAGS.output_prefix + '.infer.tar.'),
          output_kb_pattern.format(FLAGS.output_prefix + '.infer.'))
  if 'sp-train' in all_jobs:
    if FLAGS.verbose:
      print('writing self play training data')
    write_self_play(
        data, output_data_pattern.format(FLAGS.output_prefix + '.selfplay.'),
        output_kb_pattern.format(FLAGS.output_prefix + '.selfplay.'))
  if 'sp-eval' in all_jobs:
    if FLAGS.verbose:
      print('writing self play eval data')
    write_self_play(
        data,
        output_data_pattern.format(FLAGS.output_prefix + '.selfplay.eval.'),
        output_kb_pattern.format(FLAGS.output_prefix + '.selfplay.eval.'))

  if FLAGS.gen_special_token:
    # write all token file.
    f_tokens = gfile.Open(all_token_file, 'w')
    for token in list(list_of_action_tokens_except_name):
      f_tokens.write(token + '\n')
    f_tokens.close()


def run_main(unused):
  main(FLAGS)


if __name__ == '__main__':
  this_parser = argparse.ArgumentParser()
  add_arguments(this_parser)
  FLAGS, unparsed = this_parser.parse_known_args()
  # print FLAGS
  tf.app.run(main=run_main, argv=[sys.argv[0]] + unparsed)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains content related parameters."""

from tensorflow.compat.v1 import gfile


class Facts(object):
  """This class contains a collection of hyper-parameters and known facts."""

  def __init__(self, firstname_file, lastname_file, airportcode_file):
    self.airline_list = self._get_airline_list()
    self.price_limit_list = [200, 500, 1000, 5000]
    self.time_list = ['morning', 'afternoon', 'evening', 'all']
    self.time_prior = [0.03, 0.04, 0.03, 0.9]
    self.airline_preference = ['normal-cost', 'all']
    self.airline_preference_prior = [0.05, 0.95]

    # for airlines
    # mean, std
    self.flight_price_mean = {'economy': 300, 'business': 1300}
    self.flight_price_norm_std = {0: 0.2, 1: 0.4, 2: 0.6}
    self.low_cost_mean_fraction = {'economy': 0.7, 'business': 0.5}

    # prob
    self.class_member = ['economy', 'business']
    self.class_prior = [0.9, 0.1]

    self.class_list = ['all', 'economy', 'business']
    self.class_list_prior = [0.9, 0.07, 0.03]

    # prob_class={'economy':0.9,'business':0.1}
    self.max_connections = 2
    self.connection_member = [0, 1, 2]
    self.connection_prior = [0.07, 0.9, 0.03]

    # flight price= N(fraction*mean +-fraction*mean*n_std)
    self.airline_type = ['low-cost', 'normal-cost']

    self.first_name_list = self._read_file(firstname_file)
    self.last_name_list = self._read_file(lastname_file)
    airport_list_raw = self._read_file(airportcode_file)
    ob = self._get_airport_and_full_name(airport_list_raw)
    self.airport_list, self.airport_to_full_name = ob
    # all months
    self.months = [
        'Jan', 'Feb', 'Mar', 'Apr', 'May', 'June', 'July', 'Aug', 'Sept', 'Oct',
        'Nov', 'Dec'
    ]

    # has reservation
    self.has_reservation_probability = 0.10
    self.goal_probaility = [0.80, 0.1, 0.1]  # book, change, cancel
    self.goal_str_arr = ['book', 'change', 'cancel']

    # departure_time
    self.base_departure_time_epoch = 1300000000
    self.time_deviation = 3600 * 12

    self.airport_to_full_name = self._get_airline_list()

    self.final_action_list = [
        'no_flight', 'book', 'no_reservation', 'cancel', 'change'
    ]

  def _get_airline_list(self):
    """generates a list of airline and their cost attributes."""
    airline_list = {
        'UA': 'normal-cost',
        'AA': 'normal-cost',
        'Delta': 'normal-cost',
        'Hawaiian': 'normal-cost',
        'Southwest': 'low-cost',
        'Frontier': 'low-cost',
        'JetBlue': 'low-cost',
        'Spirit': 'low-cost'
    }
    return airline_list

  def _read_file(self, filename):
    original_content = gfile.Open(filename).read().strip().split('\n')
    # make sure that no empty items are added.
    content = []
    for element in original_content:
      element = element.strip()
      if element:
        content.append(element)
    return content

  def _get_airport_and_full_name(self, airport_list_raw):
    airport_to_full_name = {}
    airport_code = []
    for element in airport_list_raw:
      code, content = element.split(':')
      airport_code.append(code)
      airport_to_full_name[code] = content
    return airport_code, airport_to_full_name



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This is the setup file for airdialogue python library."""

import setuptools
with open("README.md", "r") as fh:
  long_description = fh.read()

setuptools.setup(
    name="airdialogue-essentials",
    version="0.1",
    scripts=["airdialogue/bin/airdialogue"],
    author="Wei Wei",
    author_email="wewei@google.com",
    description="airdialogue package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/javatechy/dokr",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 2",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["tqdm", "nltk", "flask"]
)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""this file contains a list of utility functions."""
import calendar
from datetime import datetime
import numpy as np


def continue_booking(facts, ag_cond, kb, goal_str):
  flights = airflight_selector(facts, ag_cond, kb)  # this is a list of flights
  if not flights:  # terminal condition
    status = 'no_flight'
  else:
    status = goal_str
  return status, flights


def airflight_selector(facts, condition, airfare_database):
  """This function selects a flight based on the condition."""
  candidates = []
  for i, flight in enumerate(airfare_database):
    cond = check_condition(facts, flight, condition, get_full_diff=False)
    if cond == 'satisfied':
      candidates.append(i)
  candidates.sort(key=lambda a: airfare_database[a]['price'])
  if not candidates:
    return None
  else:
    upper = 0
    flight = airfare_database[candidates[upper]]
    val = flight['price']
    all_good_flights = [flight]
    upper += 1
    # find all flights with the same price
    while upper < len(candidates):
      flight = airfare_database[candidates[upper]]
      if flight['price'] == val:
        all_good_flights.append(flight)
        upper += 1
      else:
        break
    return all_good_flights


def generate_expected_action(facts, agent_condition, airfare_database,
                             reservation):
  """this function generates the expected action based context."""
  goal_str = agent_condition['goal']

  if goal_str == 'book' or (goal_str == 'change' and reservation != 0):
    status, flights = continue_booking(facts, agent_condition, airfare_database,
                                       goal_str)
  elif goal_str in ['change', 'cancel'] and reservation == 0:
    status = 'no_reservation'
    flights = []
  elif goal_str == 'cancel':
    status = 'cancel'
    flights = []
  expected_action = generate_action(flights, agent_condition['name'], status)
  return expected_action


def check_states(user_action, expected_action):
  """this function check the dialogue states of the json_object.
  assume both expected action and user action went through
  action standarlization, which means they must have a flight and name field.
  """
  # first status needs to match
  status_match = expected_action['status'] == user_action['status']
  # second flight need to match
  #  if user flight is empty, expected flight has to be empty
  if len(user_action['flight']) == 0:
    flight_match = len(expected_action['flight']) == 0
  else:
    # there can only be one user flight
    assert len(user_action['flight']) == 1
    flight_match = user_action['flight'][0] in expected_action['flight']
  a = expected_action['name'].strip().lower()
  b = user_action['name'].strip().lower()
  name_match = (a == b)
  res = status_match and flight_match and name_match
  return res, status_match, flight_match, name_match


def get_day_segment_by_hour(hour):
  if hour < 3 or hour > 19:
    return 'evening'
  elif hour >= 3 and hour <= 11:
    return 'morning'
  else:
    return 'afternoon'


def check_condition(facts, flight, condition, get_full_diff=False):
  """Check the condition of the fliths to see whether it is satisfied."""
  # all the keys here have been confirmed to appear in step2 of data release.
  diff = []
  do_not_consider = set(['name', 'goal', 'departure_date', 'return_date'])
  for key in set(condition.keys()) - do_not_consider:
    if key == 'departure_time':
      day_segment = get_day_segment_by_hour(flight['departure_time_num'])
      if day_segment != condition[key]:
        diff.append(key)
    elif key == 'return_time':
      day_segment = get_day_segment_by_hour(flight['return_time_num'])
      if day_segment != condition[key]:
        diff.append(key)
    elif key == 'airline_preference':
      if condition[key] != facts.airline_list[flight['airline']]:
        diff.append(key)
    elif key == 'max_connections':
      if condition[key] < flight['num_connections']:
        # print ('ss',condition[key],flight['num_connections'])
        diff.append(key)
    elif key == 'max_price':
      if condition[key] < flight['price']:
        diff.append(key)
    else:
      # this includes, class, departure_airport/return_airport
      # departure_month/return_month, departure_day/return_day
      if flight[key] != condition[key]:
        diff.append(key)
    if not get_full_diff and len(diff):
      return diff
  if not diff:
    return 'satisfied'
  else:
    return diff


def generate_action(flights, name, status):
  """generate dialogue action."""
  action_json = {}
  if status.startswith('no_flight'):
    status = 'no_flight'
  action_json['status'] = status
  flight_lst = []  #  always have an empty flight list
  if status in ['change', 'book']:
    for f in flights:
      flight_lst.append(f['flight_number'])
  action_json['flight'] = flight_lst
  action_json['name'] = name
  # standarlize it just in case
  standarlized_action = standardize_action(action_json)
  return standarlized_action


def standardize_intent(org_intent):
  """1) get ride of departure_date and return_date in the intent.
     2) also get tide of intents with 'all'. also for max_conn if val is 2.
     3) make goal a string.
     4) replace '_' with ' ' in name.
  """
  new_intent = {}
  for key in org_intent:
    if key in ['departure_date', 'return_date']:
      continue
    if org_intent[key] == 'all':
      continue
    if key == 'max_connections' and org_intent[key] == 2:
      continue
    new_intent[key] = org_intent[key]
  new_intent['goal'] = ['book', 'change', 'cancel'][new_intent['goal']]
  new_intent['name'] = new_intent['name'].replace('_', ' ').strip()

  return new_intent


def standardize_action(org_action):
  """if status is not book or change the flight number will be empty.
  name is always required."""
  # some human raters will end a name with . or ,
  # since names in intent are standarlized (with - being replaced by space),
  # it will not be necessary to consider - again in the action standarlization.
  original_name = org_action['name'].strip()
  name = []
  for d in original_name:
    if d.isalpha() or d == ' ':
      name.append(d)
  name = ''.join(name)
  status = org_action['status']
  # if flight is a single int, we will need to convert it to a list
  # ground truth can have multiple flights
  # prediction and real data have no more than one element in the flight list.
  flight = org_action['flight']
  if type(flight) == int:
    flight = [flight]
  if status == 'book' or status == 'change':
    new_flight = []
    # get ride of anything that is not a valid flight number.
    # This could be the empty flights in early version of the UI.
    for f in flight:
      if int(f) >= 1000:
        new_flight.append(f)
  else:
    #  otherwise we provide an empty list of the flight
    #  any user selecged flights that does not come with bookable status
    #  will be ignored.
    new_flight = []

  return {'flight': new_flight, 'name': name, 'status': status}


def get_unix_epoch(dt):
  return calendar.timegm(dt.utctimetuple())


def get_datetime(unix_epoch):
  return datetime.utcfromtimestamp(unix_epoch)


def get_month_and_day(fact_obj, unix_epoch):
  m = get_datetime(unix_epoch).month
  d = get_datetime(unix_epoch).day
  month = fact_obj.months[m - 1]
  return month, str(d)


def get_hour(unix_epoch):
  return get_datetime(unix_epoch).hour


def discrete_price(original_price):
  return max(100, int(original_price / 100) * 100)


def format_time(hour):
  return str(hour) + ':00 in the ' + get_day_segment_by_hour(hour)


def get_connection(con):
  if con == 0:
    return 'direct service'
  elif con == 1:
    return '1 connection'
  else:
    return '2 connections'


def discrete_sample_(probabilities):
  if abs(sum(probabilities) - 1) > 1e-3:
    raise ValueError('sum of probability not equal to 1')
  sm = 0.0
  random_number = np.random.random()
  for i in range(len(probabilities)):
    sm += probabilities[i]
    if sm >= random_number:
      return i
  assert False, 'invalid path'


def choice(values, cnt=-1, p=None):
  if p is None:
    p = [float(1) / float(len(values))] * len(values)
  arr = []
  for _ in range(abs(cnt)):
    ind = discrete_sample_(p)
    arr.append(values[ind])
  if cnt == -1:
    return arr[0]
  else:
    return arr



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""calculates the distance between two flights."""
import numpy as np


def normalize_diff(num_a, num_b, numerical_upper, i):
  if i not in numerical_upper:
    # if no upper bound found, normalize against itself
    num_b, num_a = min(num_a, num_b), max(num_a, num_b)
    if num_b == 0.0:  #  special case when num_b is zero
      if num_a == 0.0:
        deviation = 0
      else:
        deviation = 1
    else:
      deviation = abs(num_a - num_b) * 1.0 / num_b
  else:
    deviation = abs(num_a - num_b) * 1.0 / numerical_upper[i]
    # normalize deviation to be in the same bucket from 0-1
  if deviation > 1:
    deviation = 1
  if deviation < -1:
    deviation = -1
  return abs(deviation)


def distance_calculator(flight1, flight2, debug=False):  # flight2 is benchmark
  # this will be the deviation from benchmark flight (2).
  # simple checks to deal with no flight symbol
  category_set = set([0, 1, 8, 11])
  # numerical_upper = {2: 12, 3: 12, 4: 31, 5: 31, 6: 24, 7: 24, 9: 5000, 10: 2}
  # 9 is missing
  numerical_upper = {2: 12, 3: 12, 4: 31, 5: 31, 6: 24, 7: 24, 10: 2}
  months = [
      'Jan', 'Feb', 'Mar', 'Apr', 'May', 'June', 'July', 'Aug', 'Sept', 'Oct',
      'Nov', 'Dec'
  ]
  # 0 is normal cost
  # 1 is low cost
  airline_list = {
      'UA': 0,
      'AA': 0,
      'Delta': 0,
      'Hawaiian': 0,
      'Southwest': 1,
      'Frontier': 1,
      'JetBlue': 1,
      'Spirit': 1
  }
  # total = 0.0
  values = []
  for i in range(len(flight1)):
    a = flight1[i]
    b = flight2[i]
    if i in category_set:  # for categorical features
      if i == 11:  # for airport, we need to get its attributes
        values.append(airline_list[a.split('_')[1].split('>')[0]] !=
                      airline_list[b.split('_')[1].split('>')[0]])
      else:  #  otherwise we just directly compare their categorical values
        values.append(a != b)
    else:  # numerical features
      # although 2,3 are categorical, we know they are actually numerical
      # and conver them into indices.
      if i == 2 or i == 3:
        num_a = (months.index(a.split('_')[1].split('>')[0]) + 1)
        num_b = (months.index(b.split('_')[1].split('>')[0]) + 1)
      else:
        num_a = float(a.split('_')[1].split('>')[0])
        num_b = float(b.split('_')[1].split('>')[0])
      # we calculate a deviation, which is the increment of a to b
      # compare against b
      # to make them symmetrical
      deviation = normalize_diff(num_a, num_b, numerical_upper, i)
      values.append(deviation)

      if debug:
        if i in numerical_upper:
          upper = numerical_upper[i]
        else:
          upper = None
        print(i, num_a, num_b, upper)
  if debug:
    print(values)
  return sum(values) / 12.0  # there are 12 elements


def retrieve_flight(identity, db_content):
  if 'empty' in identity:
    return None
  for flight in db_content:
    if identity == flight[-1].strip():
      return flight[0:-1]  # should not include flight number
  assert False, 'this should not happen' + str(identity) + str(type(identity))


def split_flight(actual_flight_concat):
  arr = actual_flight_concat.split('_<')
  for i in range(len(arr)):
    arr[i] = arr[i].strip()
    if arr[i][0] != '<':
      arr[i] = '<' + arr[i]
  return arr


def split_db(flight_db):
  # the first one is has reservation object
  splitted = np.array(flight_db.split(' ')[1:])
  flight_db_arr = np.reshape(splitted, [-1, 13])
  return flight_db_arr


def generate_scaled_flight(pred_idx, truth_idx_concat, db_concat):
  """generate the scaled score between two flights based on a flight distance measure.
  """
  # the first one is has reservation object
  db_content = split_db(db_concat)
  all_idx = db_content[:, -1]
  # get flight arr
  truth_idx_arr = split_flight(truth_idx_concat)
  pred_idx = pred_idx.strip()
  if '<fl_empty>' in truth_idx_arr:
    assert len(truth_idx_arr) == 1
    # this is distance measure. not equal should have maximum distance
    return pred_idx != '<fl_empty>'
  else:
    if ('empty' in pred_idx) or pred_idx not in all_idx:
      return 1  # maximum distance
    if pred_idx in truth_idx_arr:  # empty might also in truth_idx_arr
      return 0  # minimal distance

    flight_dist = []
    pred_dist = []
    for fi in all_idx:
      for fi2 in truth_idx_arr:
        if fi != fi2:
          fc = retrieve_flight(fi, db_content)
          fc2 = retrieve_flight(fi2, db_content)
          score = distance_calculator(fc, fc2)  # flight 2 is benchmark
          flight_dist.append(score)
          if fi == pred_idx:
            pred_dist.append(score)
    if not pred_dist:
      # predicted flight not in KB
      return 1
    return min(1, min(pred_dist) * 1.0 / max(flight_dist))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ROUGE metric implementation.

Copy from tf_seq2seq/seq2seq/metrics/rouge.py.
This is a modified and slightly extended verison of
https://github.com/miso-belica/sumy/blob/dev/sumy/evaluation/rouge.py.
"""






import itertools
import numpy as np


def _get_ngrams(n, text):
  """Calcualtes n-grams.

  Args:
    n: which n-grams to calculate
    text: An array of tokens

  Returns:
    A set of n-grams
  """
  ngram_set = set()
  text_length = len(text)
  max_index_ngram_start = text_length - n
  for i in range(max_index_ngram_start + 1):
    ngram_set.add(tuple(text[i:i + n]))
  return ngram_set


def _split_into_words(sentences):
  """Splits multiple sentences into words and flattens the result."""
  return list(itertools.chain(*[_.split(" ") for _ in sentences]))


def _get_word_ngrams(n, sentences):
  """Calculates word n-grams for multiple sentences."""
  assert len(sentences) > 0
  assert n > 0

  words = _split_into_words(sentences)
  return _get_ngrams(n, words)


def _len_lcs(x, y):
  """
  Returns the length of the Longest Common Subsequence between sequences x
  and y.
  Source: http://www.algorithmist.com/index.php/Longest_Common_Subsequence

  Args:
    x: sequence of words
    y: sequence of words

  Returns
    integer: Length of LCS between x and y
  """
  table = _lcs(x, y)
  n, m = len(x), len(y)
  return table[n, m]


def _lcs(x, y):
  """
  Computes the length of the longest common subsequence (lcs) between two
  strings. The implementation below uses a DP programming algorithm and runs
  in O(nm) time where n = len(x) and m = len(y).
  Source: http://www.algorithmist.com/index.php/Longest_Common_Subsequence

  Args:
    x: collection of words
    y: collection of words

  Returns:
    Table of dictionary of coord and len lcs
  """
  n, m = len(x), len(y)
  table = dict()
  for i in range(n + 1):
    for j in range(m + 1):
      if i == 0 or j == 0:
        table[i, j] = 0
      elif x[i - 1] == y[j - 1]:
        table[i, j] = table[i - 1, j - 1] + 1
      else:
        table[i, j] = max(table[i - 1, j], table[i, j - 1])
  return table


def _recon_lcs(x, y):
  """
  Returns the Longest Subsequence between x and y.
  Source: http://www.algorithmist.com/index.php/Longest_Common_Subsequence

  Args:
    x: sequence of words
    y: sequence of words

  Returns:
    sequence: LCS of x and y
  """
  i, j = len(x), len(y)
  table = _lcs(x, y)

  def _recon(i, j):
    """private recon calculation"""
    if i == 0 or j == 0:
      return []
    elif x[i - 1] == y[j - 1]:
      return _recon(i - 1, j - 1) + [(x[i - 1], i)]
    elif table[i - 1, j] > table[i, j - 1]:
      return _recon(i - 1, j)
    else:
      return _recon(i, j - 1)

  recon_tuple = tuple([x[0] for x in _recon(i, j)])
  return recon_tuple


def rouge_n(evaluated_sentences, reference_sentences, n=2):
  """
  Computes ROUGE-N of two text collections of sentences.
  Sourece: http://research.microsoft.com/en-us/um/people/cyl/download/
  papers/rouge-working-note-v1.3.1.pdf

  Args:
    evaluated_sentences: The sentences that have been picked by the summarizer
    reference_sentences: The sentences from the referene set
    n: Size of ngram.  Defaults to 2.

  Returns:
    A tuple (f1, precision, recall) for ROUGE-N

  Raises:
    ValueError: raises exception if a param has len <= 0
  """
  if len(evaluated_sentences) <= 0 or len(reference_sentences) <= 0:
    raise ValueError("Collections must contain at least 1 sentence.")

  evaluated_ngrams = _get_word_ngrams(n, evaluated_sentences)
  reference_ngrams = _get_word_ngrams(n, reference_sentences)
  reference_count = len(reference_ngrams)
  evaluated_count = len(evaluated_ngrams)

  # Gets the overlapping ngrams between evaluated and reference
  overlapping_ngrams = evaluated_ngrams.intersection(reference_ngrams)
  overlapping_count = len(overlapping_ngrams)

  # Handle edge case. This isn't mathematically correct, but it's good enough
  if evaluated_count == 0:
    precision = 0.0
  else:
    precision = overlapping_count / evaluated_count

  if reference_count == 0:
    recall = 0.0
  else:
    recall = overlapping_count / reference_count

  f1_score = 2.0 * ((precision * recall) / (precision + recall + 1e-8))

  # return overlapping_count / reference_count
  return f1_score, precision, recall


def _f_p_r_lcs(llcs, m, n):
  """
  Computes the LCS-based F-measure score
  Source: http://research.microsoft.com/en-us/um/people/cyl/download/papers/
  rouge-working-note-v1.3.1.pdf

  Args:
    llcs: Length of LCS
    m: number of words in reference summary
    n: number of words in candidate summary

  Returns:
    Float. LCS-based F-measure score
  """
  r_lcs = llcs / m
  p_lcs = llcs / n
  beta = p_lcs / (r_lcs + 1e-12)
  num = (1 + (beta**2)) * r_lcs * p_lcs
  denom = r_lcs + ((beta**2) * p_lcs)
  f_lcs = num / (denom + 1e-12)
  return f_lcs, p_lcs, r_lcs


def rouge_l_sentence_level(evaluated_sentences, reference_sentences):
  """
  Computes ROUGE-L (sentence level) of two text collections of sentences.
  http://research.microsoft.com/en-us/um/people/cyl/download/papers/
  rouge-working-note-v1.3.1.pdf

  Calculated according to:
  R_lcs = LCS(X,Y)/m
  P_lcs = LCS(X,Y)/n
  F_lcs = ((1 + beta^2)*R_lcs*P_lcs) / (R_lcs + (beta^2) * P_lcs)

  where:
  X = reference summary
  Y = Candidate summary
  m = length of reference summary
  n = length of candidate summary

  Args:
    evaluated_sentences: The sentences that have been picked by the summarizer
    reference_sentences: The sentences from the referene set

  Returns:
    A float: F_lcs

  Raises:
    ValueError: raises exception if a param has len <= 0
  """
  if len(evaluated_sentences) <= 0 or len(reference_sentences) <= 0:
    raise ValueError("Collections must contain at least 1 sentence.")
  reference_words = _split_into_words(reference_sentences)
  evaluated_words = _split_into_words(evaluated_sentences)
  m = len(reference_words)
  n = len(evaluated_words)
  lcs = _len_lcs(evaluated_words, reference_words)
  return _f_p_r_lcs(lcs, m, n)


def _union_lcs(evaluated_sentences, reference_sentence):
  """
  Returns LCS_u(r_i, C) which is the LCS score of the union longest common
  subsequence between reference sentence ri and candidate summary C. For example
  if r_i= w1 w2 w3 w4 w5, and C contains two sentences: c1 = w1 w2 w6 w7 w8 and
  c2 = w1 w3 w8 w9 w5, then the longest common subsequence of r_i and c1 is
  "w1 w2" and the longest common subsequence of r_i and c2 is "w1 w3 w5". The
  union longest common subsequence of r_i, c1, and c2 is "w1 w2 w3 w5" and
  LCS_u(r_i, C) = 4/5.

  Args:
    evaluated_sentences: The sentences that have been picked by the summarizer
    reference_sentence: One of the sentences in the reference summaries

  Returns:
    float: LCS_u(r_i, C)

  ValueError:
    Raises exception if a param has len <= 0
  """
  if len(evaluated_sentences) <= 0:
    raise ValueError("Collections must contain at least 1 sentence.")

  lcs_union = set()
  reference_words = _split_into_words([reference_sentence])
  combined_lcs_length = 0
  for eval_s in evaluated_sentences:
    evaluated_words = _split_into_words([eval_s])
    lcs = set(_recon_lcs(reference_words, evaluated_words))
    combined_lcs_length += len(lcs)
    lcs_union = lcs_union.union(lcs)

  union_lcs_count = len(lcs_union)
  union_lcs_value = union_lcs_count / combined_lcs_length
  return union_lcs_value


def rouge_l_summary_level(evaluated_sentences, reference_sentences):
  """
  Computes ROUGE-L (summary level) of two text collections of sentences.
  http://research.microsoft.com/en-us/um/people/cyl/download/papers/
  rouge-working-note-v1.3.1.pdf

  Calculated according to:
  R_lcs = SUM(1, u)[LCS<union>(r_i,C)]/m
  P_lcs = SUM(1, u)[LCS<union>(r_i,C)]/n
  F_lcs = ((1 + beta^2)*R_lcs*P_lcs) / (R_lcs + (beta^2) * P_lcs)

  where:
  SUM(i,u) = SUM from i through u
  u = number of sentences in reference summary
  C = Candidate summary made up of v sentences
  m = number of words in reference summary
  n = number of words in candidate summary

  Args:
    evaluated_sentences: The sentences that have been picked by the summarizer
    reference_sentence: One of the sentences in the reference summaries

  Returns:
    A float: F_lcs

  Raises:
    ValueError: raises exception if a param has len <= 0
  """
  if len(evaluated_sentences) <= 0 or len(reference_sentences) <= 0:
    raise ValueError("Collections must contain at least 1 sentence.")

  # total number of words in reference sentences
  m = len(_split_into_words(reference_sentences))

  # total number of words in evaluated sentences
  n = len(_split_into_words(evaluated_sentences))

  union_lcs_sum_across_all_references = 0
  for ref_s in reference_sentences:
    union_lcs_sum_across_all_references += _union_lcs(evaluated_sentences,
                                                      ref_s)
  return _f_p_r_lcs(union_lcs_sum_across_all_references, m, n)


def rouge(hypotheses, references):
  """Calculates average rouge scores for a list of hypotheses and
  references"""

  # Filter out hyps that are of 0 length
  # hyps_and_refs = zip(hypotheses, references)
  # hyps_and_refs = [_ for _ in hyps_and_refs if len(_[0]) > 0]
  # hypotheses, references = zip(*hyps_and_refs)

  # Calculate ROUGE-1 F1, precision, recall scores
  rouge_1 = [
      rouge_n([hyp], [ref], 1) for hyp, ref in zip(hypotheses, references)
  ]
  rouge_1_f, rouge_1_p, rouge_1_r = list(map(np.mean, list(zip(*rouge_1))))

  # Calculate ROUGE-2 F1, precision, recall scores
  rouge_2 = [
      rouge_n([hyp], [ref], 2) for hyp, ref in zip(hypotheses, references)
  ]
  rouge_2_f, rouge_2_p, rouge_2_r = list(map(np.mean, list(zip(*rouge_2))))

  # Calculate ROUGE-L F1, precision, recall scores
  rouge_l = [
      rouge_l_sentence_level([hyp], [ref])
      for hyp, ref in zip(hypotheses, references)
  ]
  rouge_l_f, rouge_l_p, rouge_l_r = list(map(np.mean, list(zip(*rouge_l))))

  return {
      "rouge_1/f_score": rouge_1_f,
      "rouge_1/r_score": rouge_1_r,
      "rouge_1/p_score": rouge_1_p,
      "rouge_2/f_score": rouge_2_f,
      "rouge_2/r_score": rouge_2_r,
      "rouge_2/p_score": rouge_2_p,
      "rouge_l/f_score": rouge_l_f,
      "rouge_l/r_score": rouge_l_r,
      "rouge_l/p_score": rouge_l_p,
  }
